use strict;
use Bioware::GFF 0.68;
use Tk;
use Tk::Tree;
{
    package Tk::HList;
    sub UpDown {
	my $w = shift;
	my $spec = shift;
       
	my $done = 0;
	my $anchor = $w->info('anchor');
       
	delete $w->{'shiftanchor'};
       
	unless( defined $anchor )
	 {
	  $anchor = ($w->info('children'))[0] || '';
       
	  return unless (defined($anchor) and length($anchor));
       
	  if($w->entrycget($anchor, '-state') ne 'disabled')
	   {
	    # That's a good anchor
	    $done = 1;
	   }
	  else
	   {
	    # We search for the first non-disabled entry (downward)
	    $spec = 'next';
	   }
	 }
       
	my $ent = $anchor;
        my $img= $w->_indicator_image($ent);  # plus or minus
	return if (($ent eq '#') && ($img eq 'plus'));  #special case
	# Find the prev/next non-disabled entry
	#
	while(!$done)
	{
	 $ent = $w->info($spec, $ent);
	 last unless( defined $ent );
	 next if( $w->entrycget($ent, '-state') eq 'disabled' );
	 next if( $w->info('hidden', $ent) );
	 
	 #--->our hack: does this leaf live inside a folded branch?  
	 my @ent_pieces=split(/#/,$ent);
	 shift @ent_pieces;
	 my $ent_joined;
	 my $is_inside_fold=0;
	 while (@ent_pieces) {
	    $ent_joined .= '#'. shift @ent_pieces;
	    next unless (@ent_pieces);
	    $is_inside_fold = ($w->_indicator_image($ent_joined) eq 'plus');
	    last if $is_inside_fold;
	 }
	 next if $is_inside_fold;
	 # <---
	 
	 
	 last;
	}

	
	
	unless( defined $ent )
	{
	 $w->yview('scroll', $spec eq 'prev' ? -1 : 1, 'unit');
	 return;
	}
       
	$w->anchorSet($ent);
	$w->see($ent);
       
	#if($w->cget('-selectmode') ne 'single')
	# {
	  $w->selectionClear;
	  $w->selection('set', $ent);
	  $w->Callback(-browsecmd =>$ent);
	# }
	}
    
    sub LeftRight
    {
     my $w = shift;
     my $spec = shift;
    
     delete $w->{'shiftanchor'};
    
     my $anchor = $w->info('anchor');
    
     unless(defined $anchor)
      {
       $anchor = ($w->info('children'))[0] || '';
      }
    
     my $done = 0;
     my $ent = $anchor;
    
     while(!$done)
      {
       my $e = $ent;

# ---> our hack:  We don't want to move up/down when using left/right 
#       if($spec eq 'left')
#    	{
#	 $ent = $w->info('parent', $e);
#    
#	 $ent = $w->info('prev', $e)
#	   unless(defined $ent && $w->entrycget($ent, '-state') ne 'disabled')
#	}
#       else
#	{
#	 $ent = ($w->info('children', $e))[0];
#    
#	 $ent = $w->info('next', $e)
#	   unless(defined $ent && $w->entrycget($ent, '-state') ne 'disabled')
#	}
# <---    
       last unless( defined $ent );
       last if($w->entrycget($ent, '-state') ne 'disabled');
      }
    
     unless( defined $ent )
      {
       $w->xview('scroll', $spec eq 'left' ? -1 : 1, 'unit');
       return;
      }
    
     $w->anchorSet($ent);
     $w->see($ent);
    
     #if($w->cget('-selectmode') ne 'single')
     # {
       $w->selectionClear;
       $w->selectionSet($ent);
    
       $w->Callback(-browsecmd =>$ent);
     # }
    }

    
}

use Tk::ItemStyle;
use Tk::Adjuster;
use Tk::LabFrame;
use Tk::ROText;
use Tk::DialogBox;
use Win32::FileOp;
use Win32::Clipboard;
use Win32API::Ini;
use Tk::LabEntry;

use XML::Simple;
use List::MRU;
use Unicode::String qw(latin1);


our $kgff_version='1.3.1';

#v1.0.0 07-06-05 NEW!
#v1.0.1 07-07-05 Fixed a bug that was zapping the value of Struct Fields (type=14)
#v1.0.2 07-08-05 Changed tree background to white, sorted fields alphabetically, added current filename to titlebar
#v1.0.3 07-13-05 Added Clipboard behavior, fixed Delete menu state bug, added View: Refresh, fixed some minor bugs with substrings
#v1.0.4 07-14-05 Added @ARGV check so associations and commandline parameters will be accepted
#v1.1.0 07-23-05 Saves position/size of window and adjuster widget
#                Number of structs now shown for List fields
#                Lined up beType with value fields
#                Search by Label Value StringRef or Text
#v1.1.1 07-24-05 Changed bindings of Search functions to the main window rather than the tree
#                Tree now unfolds if necessary when Search function finds an item
#                File:Exit will now save settings
#v1.1.2 07-24-05 Fixed resizing issue for WinXP users
#                Fixed bug where pasting text values resulted in a value of 0
#v1.1.3 08-15-05 Deleting a node will select the next node in the tree
#v1.1.4 10-20-05 Added MRU List, remembers maximized or minimized state in kgff.ini
#v1.1.5 11-06-05 Fixed bug with Left Frame Width=1
#v1.1.6 03-07-06 More KotOR Tool friendly - joined argument list into single argument
#v1.1.7 02-01-07 Allowed Copy/Paste of Main struct
#v1.1.8 04-12-07 Better navigation through up/down left/right keys (left/right: folds/expands  up/down: keeps selection anchor under control)
#v1.1.9 04-23-07 Added FieldType 18 - TLKRef for Jade Empire
#v1.1.10 05-19-07 Added "if (-e $temp) to mru read -- this eliminates a bug where the file no longer exists causing KGFF to croak
#v1.2.0 05-29-07 Added support for interpretation of TLK V4.0
#v1.2.1 06-06-07 Fixed bug involving strange copy/paste effects with empty Structs or empty Lists -- thanks stoffe
#v1.2.2 06-10-07 Fixed bug in GFF.pm 0.67 (now using 0.68) that was exposed by v1.2.1 -- empty Lists or Lists with single structs not being written during save
#v1.2.3 09-25-07 Small bug fix with MRU list (changed ^\d+ to ^~?\d+ in reg exp) to avoid dreaded f5 refresh necessity when a bad file was in the MRU
#                also compiled with PDK 7 for people with accented usernames
#v1.24 11-17-07 (The Witcher) Added ability to toggle between V3.3 and V3.2 language modes in View menu
#                also did some behavioral tricks with selection_breadcrumb, selection_birdie, max_fieldindex, and max_structindex so that the user does
#                not get lost when toggling language options in the middle of a big tree
#                also added ability to Go To... field index or struct index 
#v1.3.0 01-08-08 Added ability to search STRREF fields by STRREF (go figure)
#                Added ability to search STRREF fields by TEXT
#                Added ability to search CEXOLOCSTRING fields by TEXT (not just CEXLOCSUBSTRINGs)
#                Made selection background color cyan for easier viewing (why didn't I do this right from the beginning?)
#v1.3.1 02-05-08 No changes, just recompiled with PDK 7.1!
# Begin subclassing of BrowseEntry widget
#

package Tk::MyBrowseEntry;
use base qw(Tk::BrowseEntry);
Construct Tk::Widget "MyBrowseEntry";
sub LabEntryWidget { "LabEntry" }
sub ButtonWidget   { "Button"   }
sub ListboxWidget  { "Listbox"  }

sub Populate {
    if ($Tk::VERSION  < 804)  {
		my ($w, $args) = @_;

		$w->SUPER::Populate($args);

		# entry widget and arrow button
		my $lpack = delete $args->{-labelPack};
		if (not defined $lpack) {
		$lpack = [-side => 'left', -anchor => 'e'];
		}
		my $var = "";
		my $e = $w->LabEntry(-labelPack => $lpack,
				 -label => delete $args->{-label},
				 -textvariable => \$var,);
		my $b = $w->Button(-bitmap => '@' . Tk->findINC('cbxarrow.xbm'));
		$w->Advertise('entry' => $e);
		$w->Advertise('arrow' => $b);
		$b->pack(-side => 'right', -padx => 1);
		$e->pack(-side => 'right', -fill => 'x', -expand => 1, -padx => 1);

		# popup shell for listbox with values.
		my $c = $w->Toplevel(-bd => 2, -relief => 'raised');
		$c->overrideredirect(1);
		$c->withdraw;
		my $sl = $c->Scrolled( qw/Listbox -selectmode browse -scrollbars oe/ );
		$w->Advertise('choices' => $c);
		$w->Advertise('slistbox' => $sl);
		$sl->pack(-expand => 1, -fill => 'both');

		# other initializations
		$w->SetBindings;
		$w->{'popped'} = 0;
		$w->Delegates('insert' => $sl, 'delete' => $sl, get => $sl, DEFAULT => $e);
		$w->ConfigSpecs(
			-listwidth   => [qw/PASSIVE  listWidth   ListWidth/,   undef],
			-listcmd     => [qw/CALLBACK listCmd     ListCmd/,     undef],
			-browsecmd   => [qw/CALLBACK browseCmd   BrowseCmd/,   undef],
			-choices     => [qw/METHOD   choices     Choices/,     undef],
			-state       => [qw/METHOD   state       State         normal/],
			-arrowimage  => [ {-image => $b}, qw/arrowImage ArrowImage/, undef],
					#our hack...
			-varindex    => [qw/PASSIVE varindex Varindex/,undef],
		#^^^

			-variable    => '-textvariable',
		-colorstate  => [qw/PASSIVE  colorState  ColorState/,  undef],
			-command     => '-browsecmd',
			-options     => '-choices',
			DEFAULT      => [$e] );
	}

	else { #TK::VERSION >= 804
		my ($w, $args) = @_;
		$w->Tk::Frame::Populate($args);

		# entry widget and arrow button
		my $lpack = delete $args->{-labelPack};
		if (not defined $lpack) {
		$lpack = [-side => 'left', -anchor => 'e'];
		}
		my $state=delete $args->{-state};
		if (not defined $state) {
			$state='readonly';
		}
		$w->{_BE_Style} = delete $args->{-style} || $Tk::platform;
		my $Listbox  = $w->ListboxWidget;
		my $LabEntry = $w->LabEntryWidget;

		my $Button   = $w->ButtonWidget;
		my $e;
		my $var = "";
		my @LabEntry_args = (-textvariable => \$var);
		if (exists $args->{-label}) {
		$e = $w->$LabEntry(-labelPack => $lpack,
				   -label => delete $args->{-label},
				   @LabEntry_args,
				  );
		} else {
		$e = $w->$LabEntry(@LabEntry_args);
		}

		my $b = $w->$Button(-bitmap => '@' . Tk->findINC($w->{_BE_Style} eq 'MSWin32' ? 'arrowdownwin.xbm' : 'cbxarrow.xbm'));
		$w->Advertise('entry' => $e);
		$w->Advertise('arrow' => $b);

		# Pack the button to align vertically with the entry widget
		my @anch;
		my $edge = {@$lpack}->{-side};
		push(@anch,-anchor => 's') if ($edge && $edge eq 'top');
		push(@anch,-anchor => 'n') if ($edge && $edge eq 'bottom');
		$b->pack(-side => 'right', -padx => 1, @anch);

		$e->pack(-side => 'right', -fill => 'x', -expand => 1); #XXX, -padx => 1);

		# popup shell for listbox with values.
		my $c = $w->Toplevel(-bd => 2,
				 -relief => ($w->{_BE_Style} eq 'MSWin32'
						 ? "solid" : "raised"));
		$c->overrideredirect(1);
		$c->withdraw;
		my $sl = $c->Scrolled( $Listbox, qw/-selectmode browse -scrollbars oe/ );
		if ($w->{_BE_Style} eq 'MSWin32' and $Tk::platform eq 'MSWin32') {
		$sl->configure(-bg => 'SystemWindow', -relief => "flat");
		}
		$w->Advertise('choices' => $c);
		$w->Advertise('slistbox' => $sl);
		$sl->pack(-expand => 1, -fill => 'both');

		$sl->Subwidget("scrolled")->bind("<Motion>",sub {
		return unless ($w->{_BE_Style} eq 'MSWin32');
		my $e = $_[0]->XEvent;
		my $y = $e->y;
		my $inx = $sl->nearest($y);
		if (defined $inx) {
			$sl->selectionClear(0, "end");
			$sl->selectionSet($inx);
		}
	   });

		# other initializations
		$w->SetBindings;
		$w->{'_BE_popped'} = 0;
		$w->Delegates(get => $sl, DEFAULT => $e);
		$w->ConfigSpecs(
			-font        => [qw/DESCENDANTS font Font/],
			-listwidth   => [qw/PASSIVE  listWidth   ListWidth/,   undef],
			-listheight  => [{-height => $sl}, qw/listHeight ListHeight/, undef],
			-listcmd     => [qw/CALLBACK listCmd     ListCmd/,     undef],
			-autolistwidth   => [qw/PASSIVE autoListWidth AutoListWidth/, undef],
			-autolimitheight => [qw/PASSIVE autoLimitHeight AutoLimitHeight 0/],
			-browsecmd   => [qw/CALLBACK browseCmd   BrowseCmd/,   undef],
			-browse2cmd  => [qw/CALLBACK browse2Cmd  Browse2Cmd/,  undef],
			-choices     => [qw/METHOD   choices     Choices/,     undef],
			-state       => [qw/METHOD   state       State         normal/],
			-arrowimage  => [ {-image => $b}, qw/arrowImage ArrowImage/, undef],
			-variable    => [ {'-textvariable' => $e} ],
			#our hack...
			-varindex    => [qw/PASSIVE varindex Varindex/,undef],
		#^^^
			-colorstate  => [qw/PASSIVE  colorState  ColorState/,  undef],
			-command     => '-browsecmd',
			-options     => '-choices',
		-label       => [qw/PASSIVE  label       Label/,       undef],
		-labelPack   => [qw/PASSIVE  labelPack   LabelPack/,   undef],
		-buttontakefocus => [{-takefocus => $b}, 'buttonTakefocus',
					 'ButtonTakefocus', 1],
			DEFAULT      => [$e] );
	}
}
sub LbCopySelection {
	if ($Tk::VERSION < 804) {

	my ($w) = @_;
    my $index = $w->LbIndex;
	    if (defined $index) {
	$w->{'curIndex'} = $index;
	my $l = $w->Subwidget('slistbox')->Subwidget('listbox');
        my $var_ref = $w->cget( '-textvariable' );
        $$var_ref = $l->get($index);
        # our hack...
        my $var_ref2= $w->cget('-varindex');
        $$var_ref2=$index;
        #^^^^

	if ($w->{'popped'}) {
	    $w->Popdown;
	}
    }
    $w->Popdown;
	}
	else { #Tk::VERSION >=804


	my ($w) = @_;
    my $index = $w->LbIndex;

    if (defined $index) {
	$w->{'_BE_curIndex'} = $index;
	my $l = $w->Subwidget('slistbox')->Subwidget('listbox');
        my $var_ref = $w->cget( '-textvariable' );
        $$var_ref = $l->get($index);
        # our hack...
        my $var_ref2= $w->cget('-varindex');
        $$var_ref2=$index;
        #^^^^
	if ($w->{'_BE_popped'}) {
	    $w->Popdown;
	}
    }
    $w->Popdown;
	}
}

package main;

sub mnuFile;
sub mnuEdit;
sub mnuTools;
sub mnuView;



#my %node_clipboard;
my %tree_memory;
my $selection_breadcrumb; # in case of tree refreshing, we want to re-select where we were.
my $selection_birdie;     # the new treeitem to jump to after populating tree
my $current_gffname;
my $max_structindex;
my $max_fieldindex;
my $tlkfile;
my $tlkfh;
my $tlkfile_in_use;
my $auto_apply;
my $gff;
my $gff_dirty_flag;
my $mw=MainWindow->new(-title=>"K-GFF");#,-bg=>'SystemButtonFace');
my $mrulist;
my $mrulist_max=iniReadString('Settings','MRUMax','4','kgff.ini');
read_mru_list();
my @gender_types=qw(Male Female);
my %lang_v32_types=(0=>'English',
		    1=>'French',
		    2=>'German',
		    3=>'Italian',
		    4=>'Spanish',
		    5=>'Polish',
		    128=>'Korean',
		    129=>'Chinese (Trad.)',
		    130=>'Chinese (Simp.)',
		    131=>'Japanese');
my %lang_v33_types=(
0=>'Default(0)',
1=>'English(1)',
2=>'English(2)',
3=>'English(3)',
5=>'Polish(5)',
10=>'German(10)',
11=>'French(11)',
12=>'Spanish(12)',
13=>'Italian(13)',
14=>'Russian(14)',
15=>'Slovakian(15)',
16=>'Hungarian(16)',
21=>'Chienese(21)',
);

my $v32_lang=iniReadString('Main','Language ID Definition','Bioware','kgff.ini');
$v32_lang = $v32_lang eq 'Bioware' ? 'Bioware' : 'The Witcher';
iniWriteString('Main','Language ID Definition',$v32_lang,'kgff.ini');

my %language_types= $v32_lang eq 'Bioware' ? %lang_v32_types : %lang_v33_types;

my @field_types= qw(
              BYTE
              CHAR
              WORD
              SHORT
              DWORD
              INT
              DWORD64
              INT64
              FLOAT
              DOUBLE
              CEXOSTRING
              RESREF
              CEXOLOCSTRING
              BINARY
              STRUCT
              LIST
              ORIENTATION
              VECTOR
	      STRREF);

use Win32::API;
my $msgbox=Win32::API->new('user32','MessageBoxA','NPPN','N');
#   $msgbox->Call(0,"No dialog.tlk file could be found.\n"
#      ."DLGEditor cannot continue.",
#      "No Dialog.tlk files detected",0);
my $mouseclick=Win32::API->new('user32','mouse_event','NNNNN','V');
my $GetWindowPlacement=Win32::API->new('user32','GetWindowPlacement','NP','N');
my $GetLastError=Win32::API->new('kernel32','GetLastError','','N');
my $ShowWindow=Win32::API->new('user32','ShowWindow','NN','N');
my $UpdateWindow=Win32::API->new('user32','UpdateWindow','N','N');
my $IsZoomed=Win32::API->new('user32','IsZoomed','N','N');
my $IsIconic=Win32::API->new('user32','IsIconic','N','N');
my $GetVersion=Win32::API->new('kernel32','GetVersion','','N');

$mw->configure(-menu=>my $menubar=$mw->Menu(-tearoff=>0));

my $icon;
eval {  my $iconfile="drill.gif";  $icon=PerlApp::extract_bound_file($iconfile); };              #Extract .bmp from .exe
unless ($@) {
    my $image = $mw->Photo(-file => $icon, -format => 'gif');                                   #apply .bmp as application icon
    $mw->Icon(-image => $image);
}

$menubar->cascade(-label=>'~File',-menuitems=>mnuFile,-tearoff=>0);
$menubar->cascade(-label=>'~Edit',-menuitems=>mnuEdit,-tearoff=>0);
$menubar->cascade(-label=>'~Tools',-menuitems=>mnuTools,-tearoff=>0);
$menubar->cascade(-label=>'~View',-menuitems=>mnuView,-tearoff=>0);


#  -------------------
#   Left Frame (Tree)
#  -------------------

my $lfw=iniReadString('LeftFrame','Width','','kgff.ini');
$lfw=330 unless ($lfw>10);
my $lfh=iniReadString('LeftFrame','Height','','kgff.ini');
$mw->packPropagate(0);
my $left_frame= $mw->Frame(-borderwidth=>2,-relief=>'groove')->pack(-expand=>0,-fill=>'both',-side=>'left');


#  -------------------
#    Tree
#  -------------------
my $tree_bgcolor='#D7D7D7';
my $tree=$left_frame->Scrolled("Tree",
                                -scrollbars=>'se',
                                -separator=>'#',
                                -background=>'#FFFFFF',
                                -selectbackground=>'#7FFFF0',
                                -selectmode=> 'single',
				-highlightcolor=>'#00A0A0',
                                -browsecmd=>sub {my $treeitem=shift;
                                                 tree_click($treeitem);
                                                 },
                                -command=>sub { my $treeitem=shift;
                                               tree_dblclick($treeitem);
                                               }
                                )->pack(-fill=>'both',-expand=>1);


# ----------------
#   Define menus
# ----------------
sub mnuFile {
    my @mru_items;
    if ($mrulist->count) {
      @mru_items=('');
      for (my $i=1; $i<=$mrulist->count;$i++) {
          push @mru_items,['command','~'.$i." ".($mrulist->list)[$i-1],-command=>[ \&open_gff_from_arg, $mrulist->{list}[$i-1] ] ];
      }
      push @mru_items,'';
    }

    [
     ['command','~New',       -accelerator=>'Ctrl+N', -command=>sub { create_new_gff() }],
     ['command','~Open...',   -accelerator=>'Ctrl+O', -command=>sub { open_gff() }],
     ['command','~Revert',    -accelerator=>'Ctrl+Backspace', -command=>sub { open_gff_from_arg($current_gffname) },-state=>'disabled'],
     ['command','~Close',     -accelerator=>'Ctrl+F4',-command=>sub { close_gff()}, -state=>'disabled'],
     ['command','~Save',      -accelerator=>'Ctrl+S', -command=>sub { save_gff() }, -state=>'disabled'],
     ['command','Save ~As...',-accelerator=>'Ctrl+A', -command=>sub { save_as_gff () }, -state=>'disabled'],
     '',
     ['command','~Launch in GFF Editor',-accelerator=>'Ctrl+L', -command=>sub { launch_gffeditor() }],
	 ['command','Set ~MRU List Length...',-command=>sub { set_mru_length() } ],
	 @mru_items,
     ['command','E~xit',      -accelerator=>'Ctrl+Q', -command=>sub { close_gff(); $mw->destroy; exit; }]
    ]

}
sub mnuEdit {
    [
     ['command','Cu~t Node', -accelerator=>'Ctrl+X',-command=>sub { cut_node(($tree->infoSelection())[0]) }, -state=>'disabled'],
     ['command','~Copy Node', -accelerator=>'Ctrl+C',-command=>sub { copy_node( ($tree->infoSelection())[0]) }, -state=>'disabled'],
     ['command','~Paste Node', -accelerator=>'Ctrl+V',-command=>sub { paste_node(($tree->infoSelection())[0]) }, -state=>'disabled'],
	 '',
	 ['command','~Delete Node',-accelerator=>'Ctrl+Delete',-command=>sub { delete_node(($tree->infoSelection())[0]) } , -state=>'disabled'],
    ]
}
sub mnuTools {
	[
	 ['radiobutton','Use ~TLK file for StringRefs',-variable=>\$tlkfile_in_use,-value=>1,-accelerator=>'Ctrl+T',-command=>sub {use_tlk_file() }, -state=>'normal'],
	 ['radiobutton','Do ~not use TLK file for StringRefs',-variable=>\$tlkfile_in_use,-value=>0,-accelerator=>'Ctrl+Shift+T',-command=>sub {no_use_tlk_file() }, -state=>'normal'],
	 '',
	 ['command','Search by ~Label',-accelerator=>'F7',-command=>sub { search_by_label();},-state=>'disabled'],
	 ['command','Search by ~Value',-accelerator=>'F8',-command=>sub { search_by_value();},-state=>'disabled'],
	 ['command','Search by ~StringRef',-accelerator=>'F9',-command=>sub { search_by_stringref();},-state=>'disabled'],
	 ['command','Search by ~Text',-accelerator=>'F10',-command=>sub { search_by_text();},-state=>'disabled'],
         '',
         ['command','~Go to...',-accelerator=>'Ctrl+G',-command=>sub { go_to();},-state=>'disabled'],
	]
}
sub mnuView {
	[
	 ['command','~Expand All', -accelerator=>']',-command=>sub { expand_tree();}],
         ['command','~Fold All', -accelerator=>'[',-command=>sub { fold_tree();}],
	 ['command','~Refresh', -accelerator=>'F5',-command=>sub { refresh_tree();}],
         '',
         ['radiobutton','Use KotOR Language IDs',-variable=>\$v32_lang,-value=>'Bioware',-command=>sub {change_to_v32_lang()},-state=>'normal'],
	 ['radiobutton','Use The Witcher Language IDs',-variable=>\$v32_lang,-value=>'The Witcher',-command=>sub {change_to_v33_lang()},-state=>'normal'],
        ]
}


# -------------------------
#  Bind Keystrokes to Menu
# -------------------------
$mw->bind('<Control-n>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('New',-state);
   if ($state eq 'normal') { create_new_gff(); }
 });
$mw->bind('<Control-o>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Open...',-state);
   if ($state eq 'normal') { open_gff(); }
 });
$mw->bind('<Control-F4>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Close',-state);
   if ($state eq 'normal') { close_gff(); }
 });
$mw->bind('<Control-BackSpace>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Revert',-state);
   if ($state eq 'normal') { open_gff_from_arg($current_gffname) }
 });
$mw->bind('<Control-s>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Save',-state);
   if ($state eq 'normal') { save_gff(); }
 });
$mw->bind('<Control-a>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Save As...',-state);
   if ($state eq 'normal') { save_as_gff(); }
 });
$mw->bind('<Control-l>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Launch in GFF Editor',-state);
   if ($state eq 'normal') { launch_gffeditor(); }
 });
$mw->bind('<Control-q>'=>sub {
   my $file_menu=$menubar->entrycget('File',-menu);
   my $state=$file_menu->entrycget('Exit',-state);
   if ($state eq 'normal') { close_gff(); exit; }
 });


$tree->bind('<Control-x>'=>sub {
   my $edit_menu=$menubar->entrycget('Edit',-menu);
   my $state=$edit_menu->entrycget('Cut Node',-state);
   if ($state eq 'normal') { cut_node(($tree->info('selection'))[0]) }
 });
$tree->bind('<Control-c>'=>sub {
   my $edit_menu=$menubar->entrycget('Edit',-menu);
   my $state=$edit_menu->entrycget('Copy Node',-state);
   if ($state eq 'normal') { copy_node(($tree->info('selection'))[0]) }
 });
$tree->bind('<Control-v>'=>sub {
   my $edit_menu=$menubar->entrycget('Edit',-menu);
   my $state=$edit_menu->entrycget('Paste Node',-state);
   if ($state eq 'normal') { paste_node(($tree->info('selection'))[0]) }
 });
$tree->bind('<Control-Delete>'=>sub {
   my $edit_menu=$menubar->entrycget('Edit',-menu);
   my $state=$edit_menu->entrycget('Delete Node',-state);
   if ($state eq 'normal') { delete_node(($tree->info('selection'))[0]) }
 });

$tree->bind('<Key-[>'=>sub {
   my $view_menu=$menubar->entrycget('View',-menu);
   my $state=$view_menu->entrycget('Fold All',-state);
   if ($state eq 'normal') { fold_tree(); }
 });
$tree->bind('<Key-]>'=>sub {
   my $view_menu=$menubar->entrycget('View',-menu);
   my $state=$view_menu->entrycget('Expand All',-state);
   if ($state eq 'normal') { expand_tree(); }
 });
$mw->bind('<Key-F5>'=>sub {
   my $view_menu=$menubar->entrycget('View',-menu);
   my $state=$view_menu->entrycget('Refresh',-state);
   if ($state eq 'normal') { refresh_tree(); }
 });
$mw->bind('<Key-F7>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Search by Label',-state);
   if ($state eq 'normal') { search_by_label(); }
 });
$mw->bind('<Key-F8>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Search by Value',-state);
   if ($state eq 'normal') { search_by_value(); }
});
$mw->bind('<Key-F9>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Search by StringRef',-state);
   if ($state eq 'normal') { search_by_stringref(); }
});
$mw->bind('<Key-F10>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Search by Text',-state);
   if ($state eq 'normal') { search_by_text(); }
});
$mw->bind('<Control-g>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Go to...',-state);
   if ($state eq 'normal') { go_to(); }
});

$mw->bind('<Control-t>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Use TLK file for StringRefs',-state);
   if ($state eq 'normal') { use_tlk_file(); }
});
$mw->bind('<Control-T>'=>sub {
   my $tools_menu=$menubar->entrycget('Tools',-menu);
   my $state=$tools_menu->entrycget('Do not use TLK file for StringRefs',-state);
   if ($state eq 'normal') {
        no_use_tlk_file();
        $msgbox->Call(0,"TLK file will NOT be used to resolve String Refs.","Search complete",0);        
    }
});
$tree->bind('<Right>'=>sub { });
$tree->bind('<Left>'=>sub { });
$tree->bind('<Key-Right>'=>sub {
    expand_tree( ($tree->info('selection'))[0] );
    tree_click(($tree->info('selection'))[0]);
    Tk->break;
});
$tree->bind('<Key-Left>'=>sub {
    fold_tree( ($tree->info('selection'))[0] );
    tree_click(($tree->info('selection'))[0]);
    Tk->break;
});

#bind right click
$tree->bind('<ButtonPress-3>'=>sub { my $treeitem=shift; tree_rclick($treeitem); } );


#  ------------------
#   Adjuster widget
#  ------------------
my $adjuster= $mw->Adjuster(-widget=>$left_frame,-side=>'left')->pack(-side=>'left',-fill=>'both');

#  -------------------
#   Right Frame (Info)
#  -------------------
my $right_frame = $mw->Frame(-borderwidth=>2,-relief=>'groove')->pack(-expand=>1,-fill=>'both',-side=>'left');
my $fraGroup=$right_frame->Frame();
my @fieldwidgets; #widgets that should be enabled for fields, but disabled for structs
my @structwidgets; #widgets that should be enabled for structs, but disabled for fields
my @locstringwidgets; #widgets that should be enabled for CEXOLOCSTRINGS
my @locsubstringwidgets; #widgets that should be enabled for substrings of CEXOLOCSTRINGS
my @binarywidgets; #widgets that should be enabled for BINARY fields
my $lblLabel=$fraGroup->Label(-text=>'Label:');
my $txtLabel_var;
my $txtLabel=$fraGroup->Entry(-textvariable=>\$txtLabel_var,-width=>16,-state=>'disabled');

my $lblValue=$fraGroup->Label(-text=>'Value(s):');
my @txtValue_var;
my @txtValue;
for my $i (0..3) {
    $txtValue[$i]=$fraGroup->Entry(-textvariable=>\$txtValue_var[$i],-width=>16,-state=>'disabled');
}
push @fieldwidgets,(@txtValue,$txtLabel);
my $beType_var;
my $beType_varix;
#my $beType=$right_frame->MyBrowseEntry(-varindex=>\$beType_varix,-variable=>\$beType_var,-choices=>[@field_types],-label=>'Type:',-width=>16);

my $beType;
#if ($Tk::VERSION  > '800.024') {
  $beType=$fraGroup->MyBrowseEntry(-varindex=>\$beType_varix,-variable=>\$beType_var,-choices=>[@field_types],-label=>'',-width=>16);
#} else {
#  $beType=$fraGroup->BrowseEntry(-varindex=>\$beType_varix,-variable=>\$beType_var,-choices=>[@field_types],-label=>'',-width=>16);
#}
my $belbl=$beType->Subwidget('entry')->Subwidget('label');
$belbl->packForget();
my $lblType=$fraGroup->Label(-text=>'Type:');

my $beTypeEntry=$beType->Subwidget('entry')->Subwidget('entry');
eval {$beTypeEntry->configure(-disabledbackground=>'white',-disabledforeground=>'black',-state=>'disabled');};
my $lblID=$right_frame->Label(-text=>'Struct ID:');
my $fraLocStringInfo=$right_frame->Frame();
my $lblLocStringInfo=$fraLocStringInfo->Label(-text=>'(Localized strings)',-background=>'#FF00FF');
my $txtID_var;
my $txtID=$right_frame->Entry(-textvariable=>\$txtID_var,-width=>4,-state=>'disabled');
push @structwidgets,$txtID;
my $lblIndex=$right_frame->Label(-text=>'..');
my $lblStringRef=$fraGroup->Label(-text=>'StringRef:');
my $txtStringRef_var;
my $fraBinary=$right_frame->Frame();
my $btnImport=$fraBinary->Button(-text=>'Import Data',-command=>sub { import_binary() },-state=>'disabled')->pack();
my $btnExport=$fraBinary->Button(-text=>'Export Data',-command=>sub { export_binary() },-state=>'disabled')->pack();
my $lblBinarySize=$fraBinary->Label(-text=>'Bytes:');
eval {$lblBinarySize->configure(-state=>'disabled');};
$lblBinarySize->pack(-side=>'left');
push @binarywidgets,($btnImport,$btnExport,$lblBinarySize,$txtLabel);
my $fraText=$right_frame->Frame();
my $txtLocString=$fraText->Text(-background=>'white',-foreground=>'black',-height=>10,-width=>40,-state=>'disabled',-wrap=>'word');
my $txtLocStringRO=$fraText->ROText(-background=>'light gray',-foreground=>'black',-height=>10,-width=>40,-state=>'normal',-wrap=>'word')->pack();
my $txtStringRef=$fraGroup->Entry(-textvariable=>\$txtStringRef_var,-width=>6,-state=>'disabled',
                                      -validate=>'key',
                                      -validatecommand=>sub{
                                        if ($_[0]=~/^(-?\d*\.?\d*)$/) {
                                            #valid numeric input
                                            if ($_[0]==-1) {
						$lblLocStringInfo->pack();
                                            } elsif ($_[0]>-1) {
                                                $txtLocString->packForget();
						$lblLocStringInfo->packForget();
                                                $txtLocStringRO->pack();
                                                $txtLocStringRO->delete('1.0','end');
                                                $txtLocStringRO->insert("end",get_tlk_string($_[0]));
                                            }
                                            else {
						$lblLocStringInfo->packForget();
						$txtLocString->packForget();

					    }
                                            return 1;
                                        } return 0;
                                      },
                                      -invalidcommand=>sub{$mw->bell});

push @locstringwidgets,($txtStringRef,$txtLabel);
my ($beGender_varix,$beGender_var);
my $beGender=$right_frame->MyBrowseEntry(-varindex=>\$beGender_varix,
					 -variable=>\$beGender_var,
					 -choices=>[@gender_types],
					 -label=>'Gender:',
					 -width=>6);
my ($beLanguage_varix,$beLanguage_var);

my $beLanguage=$fraGroup->MyBrowseEntry(-varindex=>\$beLanguage_varix,
					   -variable=>\$beLanguage_var,
					   -choices=>[map {$language_types{$_}} sort {$a<=>$b} keys %language_types],
					   -label=>'',-width=>16);
$belbl=$beLanguage->Subwidget('entry')->Subwidget('label');
$belbl->packForget();
my $lblLanguage=$fraGroup->Label(-text=>'Language:');
my $lblTextCaption=$right_frame->Label(-text=>'CExoLocString Substring Info:');
my $lblVersion=$right_frame->Label(-text=>"v$kgff_version");
push @locsubstringwidgets,($beGender,$beLanguage,$txtLocString);
#--------------
# now grid them
#---------------

#row 0
$lblIndex->grid(   -row=>0,-column=>0,-sticky=>'w');
#row 1
#$beType->grid(     -row=>1,-column=>0,-sticky=>'e',-columnspan=>2);
$lblID->grid(      -row=>1,-column=>2,-sticky=>'e');
$txtID->grid(      -row=>1,-column=>3,-sticky=>'w');

#row 2
# fraGroup...
  # row 0
  $lblType->grid (   -row=>0,-column=>0,-sticky=>'e');
  $beType->grid(     -row=>0,-column=>1,-sticky=>'w');
  #row 1
  $lblLabel->grid(   -row=>1,-column=>0,-sticky=>'e');
  $txtLabel->grid(   -row=>1,-column=>1,-sticky=>'w');
  #row 2
  $lblValue->grid(   -row=>2,-column=>0,-sticky=>'e');
  $txtValue[0]->grid(-row=>2,-column=>1,-sticky=>'w');
  #row 3
  $txtValue[1]->grid(-row=>3,-column=>1,-sticky=>'w');
  #row 4
  $txtValue[2]->grid(-row=>4,-column=>1,-sticky=>'w');
  #row 5
  $txtValue[3]->grid(-row=>5,-column=>1,-sticky=>'w');
  #row 6
  $lblStringRef->grid(-row=>6,-column=>0,-sticky=>'e');
  $txtStringRef->grid(-row=>6,-column=>1,-sticky=>'w');
  #row 7
  $lblLanguage->grid(-row=>7,-column=>0,-sticky=>'e');
  $beLanguage->grid(-row=>7,-column=>1,-sticky=>'w');

$fraGroup->grid(-row=>1,-rowspan=>8,-column=>0,-columnspan=>2);
$fraLocStringInfo->grid(-row=>8,-column=>2,-columnspan=>2,-sticky=>'w');
#row 5
$fraBinary->grid(-row=>5,-column=>2,-columnspan=>2);
#row 8
#$beLanguage->grid(-row=>8,-column=>0,-columnspan=>2,-sticky=>'e');
$beGender->grid(-row=>8,-column=>2,-columnspan=>2,-sticky=>'w');

#row 9
$lblTextCaption->grid(-row=>9,-column=>0,-sticky=>'w');
#row 10
#$txtLocString
$fraText->grid(-row=>10,-column=>0,-columnspan=>5,-sticky=>'w');
#row 11
$lblVersion->place(-relx=>.9,-rely=>.95,-anchor=>'e');
#----------
# Styles
#----------
my $rootstyle=   $tree->ItemStyle('text',-background=>$tree_bgcolor,-foreground=>'#000000',-selectforeground=>'#000000',-selectbackground=>'#D0D0D0',-font=>['Century','8','bold']);
my $ws= $tree->ItemStyle('window');
my $lasttreeitem;

if (scalar @ARGV) { open_gff_from_arg(join ' ',@ARGV) }

$mw->bind('<Destroy>'=>sub {
my $w=$Tk::event->W;

#if (($w eq $left_frame)||($w eq $right_frame) || ($w eq $mw)) {
  #  if ($w eq $right_frame) {
#		iniWriteString('RightFrame','Width',$w->width, 'kgff.ini');
#		iniWriteString('RightFrame','Height',$w->height, 'kgff.ini');
	#}
	#els
if (($w eq $left_frame) || ($w eq $mw)) {
	if ($w eq $left_frame) {
		iniWriteString('LeftFrame','Width',$w->width, 'kgff.ini');
		iniWriteString('LeftFrame','Height',$w->height, 'kgff.ini');
	}
	if ($w eq $mw) {
		iniWriteString('Main','Geometry',$w->geometry, 'kgff.ini');
		get_window_state();
	}
}

});

#if ($lfw)  {
#  $mw->idletasks;
#  $left_frame->GeometryRequest($lfw,$lfh);
#  $mw->idletasks;
#}
#seems like for my machine anyway -- must do the main window geometry first then the pane request
my $new_geo=iniReadString('Main','Geometry','','kgff.ini');
$new_geo=~s/(\d+)x(\d+)/$1."x".($2-19)/e; #for some reason 19 pixels keep getting added to height
$new_geo='800x600' unless $new_geo;
$mw->geometry($new_geo);
if ($lfw)  {
  $mw->idletasks;
  $left_frame->GeometryRequest($lfw,$lfh);
  $mw->idletasks;
}
set_window_state();
MainLoop;



sub open_gff{
    return unless ($current_gffname=OpenDialog(title=>'Select a GFF file to open',
           filters => [ 'GFF Files' => '*.uti;*.utc;*.utd;*.gff;*.are;*.utt;*.git;*.ifo;*.utp;*.utm;*.dlg', 'All Files' => '*.*'],
           options =>OFN_FILEMUSTEXIST |  OFN_PATHMUSTEXIST | OFN_HIDEREADONLY));
    clear_tree();
    $gff=Bioware::GFF->new();
    $max_structindex=-1;
    $max_fieldindex=-1;
    $selection_breadcrumb='';
    return unless ($gff->read_gff_file($current_gffname));
	$mw->configure(-title=>"K-GFF $current_gffname");
	update_mru($current_gffname);
    populate_tree();
    $lasttreeitem=undef;
    $tree->update();
}
sub open_gff_from_arg {
    $current_gffname=shift;
	$mw->configure(-title=>"K-GFF $current_gffname");
    $gff_dirty_flag=0;
    clear_tree();
    $gff=Bioware::GFF->new();
    $max_structindex=-1;
    $max_fieldindex=-1;
    $selection_breadcrumb='';
    return unless ($gff->read_gff_file($current_gffname));
    update_mru($current_gffname);
    populate_tree();
    $tree->update();
    
}

sub save_gff{
   $gff_dirty_flag=0;
   unless ($current_gffname) { save_as_gff(); }
   if ($current_gffname) {
      #my ($sec,$min,$hour,$mday,$mon,$year,undef,undef,undef)=localtime;
      #my @months=qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
      #my $info=sprintf('v%s %s LastEdit: %2.2u-%s-%2.2u %2.2u:%2.2u:%2.2u',$version,$compilation_date,$mday,$months[$mon],$year-100,$hour,$min,$sec);
      #my $info_ix=$gff->{Main}->get_field_ix_by_label('EditorInfo');
      #if (defined $info_ix) {
      # $gff->{Main}{Fields}[$info_ix]{Value}=$info;
      #}
      #else {
      #  $gff->{Main}->createField('Type'=>FIELD_CEXOSTRING,'Label'=>'EditorInfo','Value'=>$info);
      #}
	  $current_gffname=~/\.(.{3})$/;
	  my $ext=$1;
	  $ext = 'GFF' unless $ext;
	  $ext=uc($ext).' ';
	  $gff->{sig}=$ext;
      $gff->write_gff_file($current_gffname);
	  update_mru($current_gffname);
   }
}
sub save_as_gff{ #save as dialog box
    my %parms=(
                title => "Save Output File As", handle=>0,
                filters => {  'All Files' => '*.*'},
                filename=>$current_gffname,
                options =>  OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT);
    $current_gffname = SaveAsDialog \%parms;
    if ($current_gffname) {
        save_gff();
        $tree->entryconfigure('#',-text=>$current_gffname);
		$mw->configure(-title=>"K-GFF $current_gffname");
    }
}

sub clear_tree {

    my $treeitem = '#';
    if ($tree->infoExists('#')) {
    while ($treeitem=$tree->infoNext($treeitem)) { # %tree memory will be emptied when popluate_tree is called
        $tree_memory{$treeitem}=$tree->infoHidden($treeitem);
    }}

    $tree->delete('all');
	$lasttreeitem=undef;
    #%entry_ixs=();
    #%reply_ixs=();
    #$last_node_info{nodetype}='none';
    #$last_node_info{widgetinfo}=[];
}

sub launch_gffeditor {
   if (-e 'GFFEditor.exe') { system "GFFEditor.exe \"$current_gffname\"" }
   else {
      #my $d = $mw->DialogBox(-title => "GFFEditor Not Found",-buttons => ["OK"]);
      #$d->add('Label', -text=>'GFF Editor was not found in this directory')->pack;
      #$d->Show();
      my $new_path=OpenDialog(
                           title=>'Locate GFFEditor.exe file',
                           filters=>['GFFEditor.exe'=>'GFFEditor.exe'],
                           options=>OFN_FILEMUSTEXIST|OFN_HIDEREADONLY);
      if ($new_path) { system "\"$new_path\" \"$current_gffname\"" }
   }
}
sub create_new_gff {
    $gff=Bioware::GFF->new('sig'=>'GFF ','version'=>'V3.2');
    $current_gffname=undef;
    clear_tree();
    populate_tree();
}
sub populate_tree {

    #if ($current_gffname) {
    #    $tree->add('#',-text=>"$current_gffname",-itemtype=>'text',-style=>$rootstyle);
    #}
    #else {
    #    $tree->add('#',-text=>"New Dialog",-itemtype=>'text',-style=>$rootstyle);
    #}
    $current_gffname="New GFF File" unless $current_gffname;
    $tree->add('#',-text=>"$current_gffname",-data=>$gff);
    
    populate_tree_struct('#',$gff->{Main});

	my $ti='#';
    while ($ti=$tree->infoNext($ti)) { # %tree memory will be emptied when popluate_tree is called
        $tree->show('entry',$ti) unless $tree_memory{$ti};
    }

    %tree_memory=(); #this will get set again by clear_tree
    $tree->autosetmode();
    
	# turn on some menu choices
	return unless (ref $gff eq "Bioware::GFF");
    my $file_menu=$menubar->entrycget('File',-menu);
    $file_menu->entryconfigure('Save',-state=>'normal');
    $file_menu->entryconfigure('Save As...',-state=>'normal');
    $file_menu->entryconfigure('Revert',-state=>'normal');
    $file_menu->entryconfigure('Close',-state=>'normal');
    my $tools_menu=$menubar->entrycget('Tools',-menu);
    $tools_menu->entryconfigure('Search by Label',-state=>'normal');
    $tools_menu->entryconfigure('Search by Value',-state=>'normal');
    $tools_menu->entryconfigure('Search by StringRef',-state=>'normal');
    $tools_menu->entryconfigure('Search by Text',-state=>'normal');
    $tools_menu->entryconfigure('Go to...',-state=>'normal');
}
sub populate_tree_struct {
    my ($parent_branch,$struct)=@_;
    return unless ref $struct eq "Bioware::GFF::Struct";
    my $new_branchname;
    if ($parent_branch =~/#$/) {
        $new_branchname=$parent_branch.'S'.$struct->{StructIndex}
    }
    else {
        $new_branchname=$parent_branch.'#S'.$struct->{StructIndex}
    
    }
    $max_structindex=$struct->{StructIndex} if $struct->{StructIndex}>$max_structindex;
    $struct->{ID}=unpack('l',pack('l',$struct->{ID}));
	{
	  if ($tree->infoExists($new_branchname)) {
		$new_branchname=~s/#S(\d+)$/'#S'.($1+1)/e;
		redo;
	  }
	}
    
    $tree->add($new_branchname,-text=>"[STRUCT ID: ".$struct->{ID}."]",-data=>$struct);
    $gff->{tree_lookup}{"S$struct->{StructIndex}"}=$new_branchname;
    $tree->hide('entry',$new_branchname) unless $parent_branch eq '#';
    if ($selection_breadcrumb=~/S(\d+)/) {
        if ($struct->{StructIndex}==$1) {
            $selection_birdie=$new_branchname;
        }
    }
    my $fieldsref=$struct->{Fields};
    if (ref $fieldsref eq 'Bioware::GFF::Field') {
        populate_tree_field ($new_branchname,$fieldsref);
        return;
    }
    return unless ref $fieldsref eq 'ARRAY';
    for my $f (sort { $a->{Label} cmp $b->{Label} } @$fieldsref) {
        populate_tree_field($new_branchname,$f);
    }
    return;
}
sub populate_tree_field {
    my ($parent_branch,$field)=@_;
    return unless ref $field eq "Bioware::GFF::Field";
    my $new_branchname;
    $field->{FieldIndex}=0 if $field->{FieldIndex}==undef;
    if ($parent_branch =~/#$/) {
        $new_branchname=$parent_branch.'F'.$field->{FieldIndex}
    }
    else {
        $new_branchname=$parent_branch.'#F'.$field->{FieldIndex}
    }
    $max_fieldindex=$field->{FieldIndex} if $field->{FieldIndex}>$max_fieldindex;
    my $branchtext=$field->{Label} . " [Type: " . $field_types[$field->{Type}] . "] ";
    if ($field->{Type} < 12) {
	$branchtext .= "Value: ".$field->{Value}
    }
	if ($field->{Type} == FIELD_CEXOLOCSTRING) {
		$branchtext .= "StringRef: ".$field->{Value}{StringRef}
    }
	if ($field->{Type} == FIELD_STRREF) {
		$branchtext .= "StringRef: ".$field->{Value}
	}
	elsif ($field->{Type}==FIELD_LIST) {
        my $structrefs=$field->{Value};
        if (ref $structrefs eq 'ARRAY') {
		  $branchtext .= " Structs: " . scalar @$structrefs;
		}
		elsif (ref $structrefs eq 'Bioware::GFF::Struct') {
			$branchtext .= " Structs: 1";
		}
		else {
			$branchtext .= " Structs: 0";
		}
	}
	{
	  if ($tree->infoExists($new_branchname)) {
		$new_branchname=~s/#F(\d+)$/'#F'.($1+1)/e;
		redo;
	  }
	}

    $tree->add($new_branchname,-text=>$branchtext,-data=>$field);
    $gff->{tree_lookup}{"F$field->{FieldIndex}"}=$new_branchname;
    if ($selection_breadcrumb=~/F(\d+)/) {
        if ($field->{FieldIndex}==$1) {
            $selection_birdie=$new_branchname;
        }
    }

    $tree->hide('entry',$new_branchname);
    
	if ($field->{Type} == FIELD_CEXOLOCSTRING) {
		my $substringref=$field->{Value}{Substrings};
		if (ref $substringref eq 'ARRAY') {
			my $ix=1;
			for my $ss (@$substringref) {
				my $stringid=$ss->{StringID};
				my $genderix=$stringid % 2;
				my $langid=($stringid - $genderix)/2;
				my $ss_branchname=$new_branchname."#T".$ix;
				my $ss_treetext='[LocalString '.$ix.'] ('.$gender_types[$genderix].' '.$language_types{$langid}.') ';
				if (length ($ss->{Value} > 20)) {
					$ss_treetext .= substr($ss->{Value},0,20).'...';
				}
				else {
					$ss_treetext .= $ss->{Value}
				}

				$tree->add($ss_branchname,-text=>$ss_treetext,-data=>$ss);
                                $gff->{tree_lookup}{"F$field->{FieldIndex}"."S".$ix-1}=$ss_branchname;
				$ix++;
			}
		}
    }

    if ($field->{Type}==FIELD_STRUCT) {
        populate_tree_struct($new_branchname,$field->{Value})
    }
    elsif ($field->{Type}==FIELD_LIST) {
        my $structrefs2=$field->{Value};
        if (ref $structrefs2 eq 'ARRAY') {
            for my $s (@$structrefs2) {
                populate_tree_struct($new_branchname,$s);
            }
        }
        elsif (ref $structrefs2 eq 'Bioware::GFF::Struct') {
            populate_tree_struct($new_branchname,$structrefs2);
        }
    }
}
sub tree_click {
    my $treeitem=shift;
    apply_changes();


	my $edit_menu=$menubar->entrycget('Edit',-menu);

	$edit_menu->entryconfigure('Cut Node',-state=>'disabled');
	$edit_menu->entryconfigure('Copy Node',-state=>'disabled');
	$edit_menu->entryconfigure('Paste Node',-state=>'disabled');
	$edit_menu->entryconfigure('Delete Node',-state=>'disabled');
    my $lastbranch = (split /#/,$treeitem)[-1];

	my $clip=Win32::Clipboard->new();
	my $clipdata=$clip->GetAs(CF_TEXT);
    if ($treeitem eq '#') { #root node
        if ($clipdata=~/^kgff_struct\n<opt ID="-1" StructIndex="0">/) {
            $edit_menu->entryconfigure('Paste Node',-state=>'normal')
        }
        field_click($treeitem);
    }
    elsif ($lastbranch=~/F/) {

		$edit_menu->entryconfigure('Copy Node',-state=>'normal');
		$edit_menu->entryconfigure('Cut Node',-state=>'normal');
		$edit_menu->entryconfigure('Delete Node',-state=>'normal');
		my $f=$tree->info('data',$treeitem);
		if (
		     (($clipdata=~/^kgff_struct/) &&  ($f->{Type} == FIELD_LIST))
			||
			 (($clipdata=~/^kgff_string/) &&  ($f->{Type} == FIELD_CEXOLOCSTRING))
		   ){
			   $edit_menu->entryconfigure('Paste Node',-state=>'normal')
		}
                
		field_click ($treeitem);

    }
    elsif ($lastbranch =~/S/) {
		$edit_menu->entryconfigure('Copy Node',-state=>'normal');
		$edit_menu->entryconfigure('Delete Node',-state=>'normal');
		if ($tree->infoParent($treeitem) ne '#') {$edit_menu->entryconfigure('Cut Node',-state=>'normal')}
		if ($clipdata=~/^kgff_field/) {
			$edit_menu->entryconfigure('Paste Node',-state=>'normal')
		}

		struct_click ($treeitem);
    }
	elsif ($lastbranch =~/T/) {
		$edit_menu->entryconfigure('Copy Node',-state=>'normal');
		$edit_menu->entryconfigure('Cut Node',-state=>'normal');
		$edit_menu->entryconfigure('Delete Node',-state=>'normal');
		cexolocsubstring_click($treeitem);
	}
    $lasttreeitem=$treeitem;
}
sub field_click {
    my $treeitem=shift;

    #enable field widgets and disable struct widgets
    for my $f (@structwidgets,@locsubstringwidgets,@locstringwidgets,@binarywidgets) {
	eval {$f->configure(-state=>'disabled');}
    }

    for my $f (@fieldwidgets) {
	eval {$f->configure(-state=>'normal');}
    }
    my $field=$tree->info('data',$treeitem);
    $selection_breadcrumb="F".$field->{FieldIndex};
    $beType_varix=$field->{Type};
    $beType_var=$field_types[$field->{Type}];
    $txtLabel_var=$field->{Label};
    $lblBinarySize->configure(-text=>'Bytes:');
	$txtStringRef_var=undef;
    $txtLocString->packForget();
	$txtLocStringRO->pack();


	$lblIndex->configure(-text=>'Field index: '.$field->{FieldIndex});
    if ($field->{Type}==FIELD_CEXOLOCSTRING) {
		cexolocstring_click($treeitem);
		return;
	}
	if ($field->{Type}==FIELD_BINARY) {
		binary_click($treeitem);
		return;
	}
	if ($field->{Type}==FIELD_LIST) {
		$txtValue[0]->configure(-state=>'disabled');
	}
	if ($field->{Type}==FIELD_STRREF) {
		strref_click($treeitem);
		return;
	}
	$txtStringRef_var=undef;
    if (($field->{Type}==FIELD_POSITION)||($field->{Type}==FIELD_ORIENTATION)) {
		if ($field->{Type}==FIELD_POSITION) {  #position
			for my $i (0..2) {
				$txtValue_var[$i]=$field->{Value}[$i];
			}
			$txtValue[3]->configure(-state=>'disabled');
			$txtValue_var[3]=undef;
		}
		if ($field->{Type}==FIELD_ORIENTATION) { #orientation
			for my $i (0..3) {
				$txtValue_var[$i]=$field->{Value}[$i];
			}
		}
    }
    else {
		for my $i (1..3) {
			$txtValue[$i]->configure(-state=>'disabled');
			$txtValue_var[$i]=undef;
		}
    }
    if ($field->{Type}<12) {
		$txtValue_var[0]=$field->{Value};
    }
	if ($field->{Type}==FIELD_BINARY) {
		$txtValue_var[0]=undef;
    }
    if ($field->{Type}==FIELD_STRUCT) {
		$txtValue_var[0]=undef;
    }
    if ($field->{Type}==FIELD_LIST) {
		$txtValue_var[0]=undef;
    }
    my $beLangEntry=$beLanguage->Subwidget('entry')->Subwidget('entry');
    eval{$beLangEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
    $beLanguage_var=undef;
    my $beGenderEntry=$beGender->Subwidget('entry')->Subwidget('entry');
    eval {$beGenderEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
    $beGender_var=undef;
    return;
}
sub struct_click {
    my $treeitem=shift;
    for my $f (@fieldwidgets,@locsubstringwidgets,@locstringwidgets,@binarywidgets) {
		eval{$f->configure(-state=>'disabled');}
    }
	for my $f (@structwidgets) {
		eval{$f->configure(-state=>'normal');}
	}

	$txtStringRef_var=undef;
    $txtLocString->packForget();
	$txtLocStringRO->pack();
    $lblBinarySize->configure(-text=>'Bytes:');
    my $struct=$tree->info('data',$treeitem);
    $selection_breadcrumb="S".$struct->{StructIndex};
    $beType_varix=FIELD_STRUCT;
    $beType_var=$field_types[FIELD_STRUCT];
    $txtID_var=$struct->{ID};
	$lblIndex->configure(-text=>'Struct index: '.$struct->{StructIndex});
	for my $i (0..3) {
		$txtValue_var[$i]=undef;
	}
	$txtLabel_var=undef;

	my $beLangEntry=$beLanguage->Subwidget('entry')->Subwidget('entry');
    eval {$beLangEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
    $beLanguage_var=undef;
	my $beGenderEntry=$beGender->Subwidget('entry')->Subwidget('entry');
	eval {$beGenderEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
    $beGender_var=undef;
}
sub strref_click {
    my $treeitem=shift;
    for my $w(@structwidgets,@fieldwidgets,@locsubstringwidgets,@binarywidgets) {
	    eval{$w->configure(-state=>'disabled');};
    }
    for my $w (@locstringwidgets) {
	eval{$w->configure(-state=>'normal');};
    }
    for my $i (0..3) {
		$txtValue_var[$i]=undef;
	}
    my $field=$tree->info('data',$treeitem);
    $txtStringRef_var=$field->{Value};
    $txtLocString->packForget();
    $txtLocStringRO->pack();
    $txtLocStringRO->delete('1.0','end');
    $txtLocStringRO->insert('end',get_tlk_string($txtStringRef_var));
}
sub cexolocstring_click {
	my $treeitem=shift;
    for my $w (@structwidgets,@fieldwidgets,@locsubstringwidgets,@binarywidgets) {
	    eval{$w->configure(-state=>'disabled');};
    }
	for my $w (@locstringwidgets) {
		eval{$w->configure(-state=>'normal');};
	}
	my $field=$tree->info('data',$treeitem);
	$txtStringRef_var=$field->{Value}{StringRef};
    if ($txtStringRef_var ==-1) { $lblLocStringInfo->pack(); }
	$txtLocString->packForget();
	$txtLocStringRO->pack();
	$txtLocStringRO->delete('1.0','end');
	$txtLocStringRO->insert('end',get_tlk_string($txtStringRef_var));
	my $beLangEntry=$beLanguage->Subwidget('entry')->Subwidget('entry');
    eval {$beLangEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
	$beLanguage_var=undef;
	my $beGenderEntry=$beGender->Subwidget('entry')->Subwidget('entry');
	eval {$beGenderEntry->configure(-disabledbackground=>'light gray',-disabledforeground=>'black',-state=>'disabled');};
	$beGender_var=undef;
}
sub cexolocsubstring_click {
	my $treeitem=shift;
    for my $f (@structwidgets,@fieldwidgets,@locstringwidgets,@binarywidgets) {
		eval{$f->configure(-state=>'disabled');};
    }
	for my $f (@locsubstringwidgets) {
		eval{$f->configure(-state=>'normal');};
	}
	$lblBinarySize->configure(-text=>'Bytes:');
	my $ssid=(split /#/,$treeitem)[-1];
	$ssid=~s/T//;
	$lblIndex->configure(-text=>'Substring '.$ssid);
    my $beLangEntry=$beLanguage->Subwidget('entry')->Subwidget('entry');
    eval{ $beLangEntry->configure(-disabledbackground=>'white',-disabledforeground=>'black',-state=>'disabled');};

	my $beGenderEntry=$beGender->Subwidget('entry')->Subwidget('entry');
	eval {$beGenderEntry->configure(-disabledbackground=>'white',-disabledforeground=>'black',-state=>'disabled');};

	my $ss=$tree->info('data',$treeitem);
	my $stringid=$ss->{StringID};
	my $genderix=$stringid % 2;
	my $langid=($stringid - $genderix)/2;
	$beGender_varix=$genderix;
	$beGender_var=$gender_types[$genderix];
	$beLanguage_varix=$langid;
	$beLanguage_var=$language_types{$langid};
	$txtLocString->delete('1.0','end');
	$txtLocString->insert('end',$ss->{Value});
	$txtLocStringRO->packForget();
	$txtLocString->pack();
}
sub binary_click {
	my $treeitem=shift;
	for my $f (@structwidgets,@locsubstringwidgets,@locstringwidgets,@fieldwidgets) {
		eval {$f->configure(-state=>'disabled');}
    }
	for my $f (@binarywidgets) {
		eval {$f->configure(-state=>'normal');}
	}
	my $field=$tree->info('data',$treeitem);
    my $binlength=do {use bytes; length ($field->{Value})};
	$lblBinarySize->configure(-text=>'Bytes: '.$binlength);
}
sub tree_rclick {
    #simulate a left-click
    my $w=shift;
    my $ev = $w->XEvent();
    $tree->Button1($ev);
    $tree->ButtonRelease1($ev);
    my $treeitem=($tree->info('selection'))[0]; #get the item name selected
    #now perform right-click actions...
    #create context menu
    my $context_menu = $mw->Menu(-tearoff  => 0);
	$context_menu->add('command',-label=>'New',-accelerator=>'Ctrl+N',-command=>sub{ create_new_gff();simclick(); });
    $context_menu->add('command',-label=>'Open',-accelerator=>'Ctrl+O',-command=>sub{open_gff();simclick(); });
	if ($current_gffname) {
		$context_menu->add('command',-label=>'Close',-accelerator=>'Ctrl+F4',-command=>sub{ close_gff();simclick(); });
	}
	my $clip=Win32::Clipboard->new();
	my $clipdata=$clip->GetAs(CF_TEXT);

    if ($treeitem eq '') {     # empty tree
    }
	else {
		if ($treeitem eq '#') { #root selected
		    my $state=$clipdata=~/^kgff_struct\n<opt ID="-1" StructIndex="0">/ ? 'normal' : 'disabled';
		    $context_menu->add('command',-label=>'Paste MAIN STRUCT',-accelerator=>'Ctrl+V', -command=>sub { paste_node($treeitem);simclick(); }, -state=>$state);
		}
		elsif ($treeitem =~/#S\d+$/) { #struct selected


			my $edit_menu=$menubar->entrycget('Edit',-menu);
			my $state=$edit_menu->entrycget('Cut Node',-state);
			my $i;
			for ($i=0; $i<scalar @field_types; $i++) {
			    my $j=$i;
				$context_menu->add('command',-label=>'Add Field: '.$field_types[$i],-command=>sub {add_field_node($treeitem,$j);simclick();},-state=>'normal');
			}
			$context_menu->add('separator');
			$state=$tree->infoParent($treeitem) eq '#' ? 'disabled' : 'normal';
            $context_menu->add('command',-label=>'Cut STRUCT',-accelerator=>'Ctrl+X',-command=>sub { cut_node($treeitem);simclick(); },-state=>$state);
            $context_menu->add('command',-label=>'Copy STRUCT',-accelerator=>'Ctrl+C',-command=>sub { copy_node($treeitem);simclick();},-state=>'normal');
            #$state=$node_clipboard{type} eq 'field' ? 'normal' : 'disabled';
			$state=$clipdata=~/^kgff_field/ ? 'normal' : 'disabled';
            $context_menu->add('command',-label=>'Paste FIELD',-accelerator=>'Ctrl+V', -command=>sub { paste_node($treeitem); simclick(); }, -state=>$state);
			$context_menu->add('separator');
			$context_menu->add('command',-label=>'Delete Struct',-accelerator=>'Ctrl+Del',-command=>sub {
				my $next_treeitem=$treeitem;
				{
				$next_treeitem=$tree->infoNext($next_treeitem);
				redo if $next_treeitem=~/^$treeitem#/; #we don't want children
				}
				$next_treeitem=$tree->infoPrev($treeitem) unless $next_treeitem;
				delete_struct_node($treeitem); $lasttreeitem=undef; simclick();
				#now select next node
				return unless $tree->infoExists($next_treeitem);
				$tree->selectionClear();
				$tree->selectionSet($next_treeitem);
			    #now see the node
			    $tree->see($next_treeitem);
				#spawn widgets ala left click
				tree_click($next_treeitem);
			}, -state=>'normal');
	#		$context_menu->add('separator');
#			$context_menu->add('command',-label=>'Move Up',-command=>sub { move_node_up{$treeitem};simclick(); }, -state=>'disabled');
#            $context_menu->add('command',-label=>'Move Down',-command=>sub { move_node_down{$treeitem};simclick();  }, -state=>'disabled');

		}
		elsif ($treeitem =~/#F\d+$/) { #field selected
			my $field=$tree->info('data',$treeitem);
			$context_menu->add('command',-label=>'Cut FIELD',-accelerator=>'Ctrl+X',-command=>sub { cut_node($treeitem);simclick(); },-state=>'normal');
            $context_menu->add('command',-label=>'Copy FIELD',-accelerator=>'Ctrl+C',-command=>sub { copy_node($treeitem);simclick(); },-state=>'normal');

			if ($field->{Type} == FIELD_CEXOLOCSTRING) {
				$context_menu->add('command',-label=>'Add String',-command=>sub { add_substring_node($treeitem);simclick(); }, -state=>'normal');
				#my $state=$node_clipboard{type} eq 'string' ? 'normal' : 'disabled';
				my $state=$clipdata=~/^kgff_string/ ? 'normal' : 'disabled';
				$context_menu->add('command',-label=>'Paste String',-accelerator=>'Ctrl+V', -command=>sub { paste_node($treeitem);simclick();}, -state=>$state);
			}
			elsif ($field->{Type} == FIELD_LIST) {
				$context_menu->add('command',-label=>'Add STRUCT',-command=>sub { add_struct_node($treeitem);simclick(); }, -state=>'normal');
				#my $state=$node_clipboard{type} eq 'struct' ? 'normal' : 'disabled';
				my $state=$clipdata=~/^kgff_struct/ ? 'normal' : 'disabled';
				$context_menu->add('command',-label=>'Paste STRUCT',-accelerator=>'Ctrl+V', -command=>sub { paste_node($treeitem);simclick(); }, -state=>$state);
			}
			$context_menu->add('separator');
			$context_menu->add('command',-label=>'Delete Field',-accelerator=>'Ctrl+Del',-command=>sub {
				my $next_treeitem=$treeitem;
				{
				$next_treeitem=$tree->infoNext($next_treeitem);
				redo if $next_treeitem=~/^$treeitem#/; #we don't want children
				}
				$next_treeitem=$tree->infoPrev($treeitem) unless $next_treeitem;

				delete_field_node($treeitem); $lasttreeitem=undef; simclick();
				return unless $tree->infoExists($next_treeitem);
				#now select next node
				$tree->selectionClear();
				$tree->selectionSet($next_treeitem);
			    #now see the node
			    $tree->see($next_treeitem);
				#spawn widgets ala left click
				tree_click($next_treeitem);

			}, -state=>'normal');
		}
		elsif ($treeitem =~/#T\d+$/) { #substring selected
			$context_menu->add('command',-label=>'Cut String',-accelerator=>'Ctrl+X',-command=>sub { cut_node($treeitem);simclick(); },-state=>'normal');
            $context_menu->add('command',-label=>'Copy String',-accelerator=>'Ctrl+C',-command=>sub { copy_node($treeitem);simclick(); },-state=>'normal');
			$context_menu->add('command',-label=>'Delete String',-accelerator=>'Ctrl+Del',-command=>sub {
				my $next_treeitem=$treeitem;
				{
				$next_treeitem=$tree->infoNext($next_treeitem);
				redo if $next_treeitem=~/^$treeitem#/; #we don't want children
				}
				$next_treeitem=$tree->infoPrev($treeitem) unless $next_treeitem;

				delete_substring_node($treeitem); $lasttreeitem=undef;simclick();
				return unless $tree->infoExists($next_treeitem);
				#now select next node
				$tree->selectionClear();
				$tree->selectionSet($next_treeitem);
			    #now see the node
			    $tree->see($next_treeitem);
				#spawn widgets ala left click
				tree_click($next_treeitem);

			}, -state=>'normal');
		}
	}
	$context_menu->Popup(
    -popover  => 'cursor',
    -popanchor => 'nw');


}
sub add_struct_node { #add a struct node to a list field
	my $treeitem=shift;
	my $dataitem=$tree->info('data',$treeitem);
    return unless ((ref $dataitem eq 'Bioware::GFF::Field') && ($dataitem->{Type} == FIELD_LIST));
	$max_structindex++;
        my $new_struct=Bioware::GFF::Struct->new('ID'=>0,'StructIndex'=>$max_structindex);
	push @{$dataitem->{Value}},$new_struct;
	my @children=$tree->info('children',$treeitem);
	my $lastchild=$children[-1];
	$lastchild=~/#S(\d+)$/;
	my $newbranch=$treeitem.'#S'.($1+1);
	$tree->add($newbranch,-text=>"[STRUCT ID: 0]",-data=>$new_struct);
	$tree->autosetmode;
	$tree->entryconfigure($treeitem,-text=>$dataitem->{Label} . " [Type: LIST] Structs: " . scalar @{$dataitem->{Value}});

}
sub add_field_node {  #add a field node to a struct
	my ($treeitem,$node_type) =@_;
	my $dataitem=$tree->info('data',$treeitem);
	return unless (ref $dataitem eq 'Bioware::GFF::Struct');

	$dataitem->createField('Type'=>$node_type,'Label'=>'New '.$field_types[$node_type]);

	my $fieldsref=$dataitem->{Fields};
	my $new_field;
	if (ref $fieldsref eq 'Bioware::GFF::Field') {
		$new_field=$dataitem->{Fields};
	}
	elsif (ref $fieldsref eq 'ARRAY') {
		$new_field=$$fieldsref[-1];
	}
        $max_fieldindex++;
	$new_field->{FieldIndex}=$max_fieldindex;
	my @children=$tree->info('children',$treeitem);
	my $lastchild=$children[-1];
	$lastchild=~/.*#F(\d+)$/;
	my $newix=$1+1;
	while ($tree->infoExists($treeitem.'#F'.$newix)) { $newix++;}
	my $newbranch=$treeitem.'#F'.$newix;
	my $branchtext=$new_field->{Label} . ' [Type: ' . $field_types[$node_type] . '] ';
	if ($node_type < 12) {
		$branchtext .= "Value: 0";
		$new_field->{Value}=0;
	}
	if ($node_type == FIELD_CEXOLOCSTRING) {
		$branchtext .= "StringRef: 0";
		$new_field->{Value}{StringRef}=0;
	}
        if ($node_type == FIELD_STRREF) {
		$branchtext .= "StringRef: 0";
		$new_field->{Value}=0;
	}
	$tree->add($newbranch,-text=>$branchtext,-data=>$new_field);
	$tree->autosetmode;
        $tree->anchorClear();
        $tree->selectionClear();                              #clear previous selection
        $mw->idletasks;
        $tree->anchorSet($newbranch);
        $tree->selectionSet($newbranch);                       #set new selection
        $mw->idletasks;
        $tree->see($newbranch);                                #make it visible
        tree_click($newbranch);                                #click it
        
}
sub add_substring_node {
	my $treeitem=shift;
	my $dataitem=$tree->info('data',$treeitem);
    return unless ((ref $dataitem eq 'Bioware::GFF::Field') && ($dataitem->{Type} == FIELD_CEXOLOCSTRING));
	my $new_substring=Bioware::GFF::CExoLocSubString->new('StringID'=>0,'Value'=>'');
	if (ref $dataitem->{Value}{Substrings}[0] eq 'Bioware::GFF::CExoLocSubString') {
		push @{$dataitem->{Value}{Substrings}},$new_substring;
	}
	else {
		$dataitem->{Value}{Substrings}=[$new_substring];
	}
	my @children=$tree->info('children',$treeitem);
	my $lastchild=$children[-1];
	my $newbranch;
	my $newix=1;
	if ($lastchild=~/.*#T(\d+)$/) {
	  $newix=$1+1;
	}
	$newbranch=$treeitem.'#T'.$newix;
	my $branchtext='[LocalString '.$newix.'] ('.$gender_types[0].' '.$language_types{0}.')';
	$tree->add($newbranch,-text=>$branchtext,-data=>$new_substring);
	$tree->autosetmode;

}
sub delete_struct_node {
	my $treeitem=shift;
	my $parent=$tree->infoParent($treeitem);
	my @siblings=$tree->infoChildren($parent);

	if ($parent ne '#') {
		my $parentdata=$tree->info('data',$parent);
		my @newstructarray;
		for my $s (@siblings) {
			unless ($s eq $treeitem) {
				push @newstructarray,$tree->info('data',$s);
			}
		}
		$parentdata->{Value}=[@newstructarray];
		$tree->delete('entry',$treeitem);
		$lasttreeitem=undef;
	}

}
sub delete_field_node {
	my $treeitem=shift;
	my $parent=$tree->infoParent($treeitem);
	my @siblings=$tree->infoChildren($parent);
	my $parentdata=$tree->info('data',$parent);
	my @newfieldarray;
	for my $s (@siblings) {
		unless ($s eq $treeitem) {
			push @newfieldarray,$tree->info('data',$s);
		}
	}
	if (scalar @newfieldarray == 1) {
		$parentdata->{Fields}=$newfieldarray[0];
	}
	else {
		$parentdata->{Fields}=[@newfieldarray];
	}
	$tree->delete('entry',$treeitem);
	$lasttreeitem=undef;
}
sub delete_substring_node {
	my $treeitem=shift;
	my $parent=$tree->infoParent($treeitem);
	my @siblings=$tree->infoChildren($parent);
	my $parentdata=$tree->info('data',$parent);
	my @newstringarray;
	for my $s (@siblings) {
		unless ($s eq $treeitem) {
			push @newstringarray,$tree->info('data',$s);
		}
	}
	$parentdata->{Value}{Substrings}=[@newstringarray];
	$tree->delete('entry',$treeitem);
	$lasttreeitem=undef;
}
sub tree_dblclick {
}
sub use_tlk_file {
    return unless ($tlkfile=OpenDialog(title=>'Select a TLK file to open',
           filters => [ 'TLK Files' => '*.tlk', 'All Files' => '*.*'],
           options =>OFN_FILEMUSTEXIST |  OFN_PATHMUSTEXIST | OFN_HIDEREADONLY));
	open $tlkfh,"<",$tlkfile;
	binmode $tlkfh;
	$tlkfile_in_use=1;
}
sub no_use_tlk_file {
	close $tlkfh;
	$tlkfile_in_use=0;
}
sub get_tlk_string {
    my $resrefnum=shift;
	return "" unless $tlkfile_in_use;
    return "" if $resrefnum==-1;
	return "" if $resrefnum==undef;
    
    seek $tlkfh,0,0;
    read $tlkfh,(my $tlkversion),8;
    if ($tlkversion eq 'TLK V4.0') {
        return get_tlk4_string($resrefnum);
    }
    seek $tlkfh,12,0;
    read $tlkfh,(my $string_count_packed),4;
	my $string_count=unpack('V',$string_count_packed);

    return '[Out of TLK file range]' if ($resrefnum>$string_count);
	return '[Bad StringRef]' if ($resrefnum<=0);
	read $tlkfh,(my $offset_packed),4;
    my $offset=unpack('V',$offset_packed);
    seek $tlkfh, 20+(40*$resrefnum)+28,0;
    read $tlkfh,(my $info_packed),8;
    my ($offset2,$size)=unpack('V2',$info_packed);
    seek $tlkfh,$offset+$offset2,0;
    read $tlkfh,(my $string),$size;
    return $string;
}
sub get_tlk4_string {
    my $resrefnum=shift;
    seek $tlkfh,12,0;
    read $tlkfh,(my $string_count_packed),4;
    my $string_count=unpack('V',$string_count_packed);
    return '[Out of TLK file range]' if ($resrefnum>$string_count);
    return '[Bad StringRef]' if ($resrefnum<=0);
    seek $tlkfh,32+(10*$resrefnum)+4,0;
    read $tlkfh,(my $info_packed),6;
    my ($offset,$size)=unpack('Vv',$info_packed);
    seek $tlkfh,$offset,0;
    read $tlkfh,(my $string),$size;
    return $string;

}

sub close_gff{
    #last_chance_to_save();
    clear_tree();
    $gff=undef;
	my $file_menu=$menubar->entrycget('File',-menu);
    $file_menu->entryconfigure('Save',-state=>'disabled');
    $file_menu->entryconfigure('Save As...',-state=>'disabled');
	$file_menu->entryconfigure('Revert',-state=>'disabled');
    $file_menu->entryconfigure('Close',-state=>'disabled');
	my $edit_menu=$menubar->entrycget('Edit',-menu);
	$edit_menu->entryconfigure('Cut Node',-state=>'disabled');
	$edit_menu->entryconfigure('Copy Node',-state=>'disabled');
	$edit_menu->entryconfigure('Paste Node',-state=>'disabled');
	$edit_menu->entryconfigure('Delete Node',-state=>'disabled');
	my $tools_menu=$menubar->entrycget('Tools',-menu);
	$tools_menu->entryconfigure('Search by Label',-state=>'disabled');
	$tools_menu->entryconfigure('Search by Value',-state=>'disabled');
	$tools_menu->entryconfigure('Search by StringRef',-state=>'disabled');
	$tools_menu->entryconfigure('Search by Text',-state=>'disabled');
	$mw->configure(-title=>"K-GFF");
}
sub cut_node {
	my $treeitem=shift;
	copy_node($treeitem);
	delete_node($treeitem);
}
sub copy_node{
	my $treeitem=shift;
	my $dataitem=$tree->info('data',$treeitem);
	#$node_clipboard{data}=$dataitem;
	my $xs=XML::Simple->new();
	my $ungff;
	$ungff=unref_gff1(\$dataitem,$ungff);
	if ((exists $ungff->{Value}) && ($ungff->{Value}==undef) && ($ungff->{Value} eq '')) {
	  $ungff->{Value}=0;
	}
        if (ref ($ungff->{Value}) eq 'ARRAY') {
            if (scalar @{$ungff->{Value}} == 0) {
                $ungff->{Value}=0;
            }
        }
        if (ref ($ungff->{Fields}) eq 'ARRAY') {
            if (scalar @{$ungff->{Fields}} == 0) {
                $ungff->{Fields}=0;
            }
        }
	if (ref $dataitem eq 'Bioware::GFF::CExoLocSubString') { #handle line breaks...
	  $ungff->{Value} =~ s/\n/&&/g;
	}
	my $xml=$xs->XMLout($ungff);
    my $clip;

	if (ref $dataitem eq 'Bioware::GFF::Struct'){
		$clip = Win32::Clipboard("kgff_struct\n$xml");
		#$node_clipboard{type}='struct';
	}
	elsif (ref $dataitem eq 'Bioware::GFF::Field') {
		$clip = Win32::Clipboard("kgff_field\n$xml");
		#$node_clipboard{type}='field';
	}
	elsif (ref $dataitem eq 'Bioware::GFF::CExoLocSubString') {
		$clip = Win32::Clipboard("kgff_string\n$xml");
		#$node_clipboard{type}='string';
	}
}
sub paste_node {
	my $treeitem=shift;
	my $dataitem=$tree->info('data',$treeitem);

	my $clip=Win32::Clipboard->new();
	my $clipdata=$clip->Get();
	$clipdata=~s/^kgff_(.*?)\n//;
    my $type=$1;
	my $xs=XML::Simple->new();
	$clipdata=$xs->XMLin($clipdata);


	my @children=$tree->infoChildren($treeitem);

	
	if ($treeitem eq '#') { # pasting the MAIN struct
            my $new_main_struct;
            $new_main_struct=unref_gff1(\$clipdata,$new_main_struct);
            $new_main_struct=reref_gff_struct($new_main_struct);
            $gff->{Main}=$new_main_struct;
        }
	elsif ($type  eq 'struct') { #pasting a struct into a list
	    my $new_uncopy;
		$new_uncopy=unref_gff1(\$clipdata,$new_uncopy);
		my $new_copy=reref_gff_struct($new_uncopy);
		push @{$dataitem->{Value}},$new_copy;
	}
	elsif ($type eq 'field') { #pasting a field into a struct
	    my $new_uncopy;
		$new_uncopy=unref_gff1(\$clipdata,$new_uncopy);
		my $new_copy=reref_gff_field($new_uncopy);
		if (ref $dataitem->{Fields} eq 'Bioware::GFF::Field') {
			$dataitem->{Fields}=[$dataitem->{Fields}, $new_copy];
		}
		elsif (ref $dataitem->{Fields} eq 'ARRAY') {
			push @{$dataitem->{Fields}},$new_copy;
		}
		else {
			$dataitem->{Fields}=$new_copy;
		}
	}

	elsif ($type eq 'string') { #pasting a CExoLocSubString into a CExoLocString
	    my $new_uncopy;
		$new_uncopy=unref_gff1(\$clipdata,$new_uncopy);
		$new_uncopy->{Value}=~s/&&/\n/g;
		my $new_copy=reref_gff_substring($new_uncopy);
		push @{$dataitem->{Value}{Substrings}},$new_copy;
	}
	refresh_tree();

}
sub refresh_tree {
	clear_tree();
	populate_tree();
        if ($tree->infoExists($selection_birdie)) {
            $tree->selectionSet($selection_birdie);
            $mw->idletasks;
            $tree->see($selection_birdie);
            tree_click($selection_birdie);
        }

}











sub simclick {
$mouseclick->Call(2,0,0,0,0);
$mouseclick->Call(4,0,0,0,0);
}
sub import_binary {
    # need to get the treeitem that is selected
	my $treeitem=($tree->info('selection'))[0];
	return unless $treeitem;
	my $field=$tree->info('data',$treeitem);
	return unless ( (ref $field eq 'Bioware::GFF::Field') &&
				    ($field->{Type}==FIELD_BINARY) );
	my $fn;
	my $fh;
    return unless ($fn=OpenDialog(title=>'Select a file to store inside BINARY FIELD',
           filters => [ 'All Files' => '*.*'],
           options =>OFN_FILEMUSTEXIST |  OFN_PATHMUSTEXIST | OFN_HIDEREADONLY));

	local $/;
	open $fh,"<",$fn;
	binmode $fh;
	my $data=<$fh>;
	close $fh;
    $field->{Value}=$data;
	my $binsize=do {use bytes; length $field->{Value}};
	$lblBinarySize->configure(-text=>'Bytes: '.$binsize);
	$msgbox->Call(0,"Data imported ($binsize bytes)","Import successful",0);
	return;
}
sub export_binary {
    # need to get the treeitem that is selected
	my $treeitem=($tree->info('selection'))[0];
	return unless $treeitem;
	my $field=$tree->info('data',$treeitem);
	return unless ( (ref $field eq 'Bioware::GFF::Field') &&
				    ($field->{Type}==FIELD_BINARY) );
	my $fn;
	my $fh;
    return unless ($fn=SaveAsDialog(title=>'Select a file for export of BINARY FIELD',
           filters => [ 'All Files' => '*.*'],
           options =>  OFN_OVERWRITEPROMPT| OFN_PATHMUSTEXIST));
	open $fh,">",$fn;
	binmode $fh;
	syswrite $fh,$field->{Value};
	close $fh;
	my $binsize=do{use bytes; length $field->{Value}};
	$lblBinarySize->configure(-text=>"Bytes: $binsize");
	$msgbox->Call(0,"Data exported ($binsize bytes)","Export successful",0);
}
sub unref_gff1{ #this recursive subroutine derefences the gff object
  my $g_ref=shift; #the gff reference
  my $h=shift; #the ungff
  my $g=$$g_ref;  #the gff
  $h=unref_gff2($g); #drop the class and return a hash ref, array ref, or scalar
  if ((ref $g)=~/Bioware/) { #for classes, go through each key and value
    for my $gk (keys %$g) {
      my $hk;
      $hk=unref_gff1(\$gk,$hk); #drop any classes in the keys
      my $hv;
      my $gv=$g->{$gk};
      $hv=unref_gff1(\$gv,$hv); #drop any classes in the values
      $h->{$hk}=$hv; #create our unref'd gff
    }
  }
  elsif (ref $g eq 'ARRAY') { #for arrays, go through each element
    for (my $i=0; $i<scalar @$g; $i++) {
      my $gix=$g->[$i];
      my $hix;
      $hix=unref_gff1(\$gix,$hix);  #drop classes for each element
      $h->[$i]=$hix; #create our unref'd gff
    }
  }
  return $h;

}


sub unref_gff2{ #returns hash ref, array ref or scalar
 my $g = shift;
  if ((ref $g) =~ /Bioware/) {
    return {%$g};
  }
  elsif ((ref $g) eq 'ARRAY') {
    return [@$g];
  }
  else {
    return $g
  }
}
sub reref_gff{
  my $ungff=shift;
  my $new_gff=Bioware::GFF->new();
  $new_gff->{sig}=$ungff->{sig};
  $new_gff->{version}=$ungff->{version};
  my $ungff_main_struct=$ungff->{Main};
  $new_gff->{Main}=reref_gff_struct($ungff_main_struct);
  return $new_gff;
}
sub reref_gff_struct {
  my $ungff_struct=shift;
  my $gff_struct=Bioware::GFF::Struct->new();
  $gff_struct->{ID}=$ungff_struct->{ID};
  $gff_struct->{StructIndex}=$ungff_struct->{StructIndex};
  if (ref $ungff_struct->{Fields} eq 'ARRAY') {
	for my $ungff_field (@{$ungff_struct->{Fields}}) {
	 my $gff_field=reref_gff_field($ungff_field);
	 unless (ref $gff_struct->{Fields} eq 'ARRAY') {
	    if ($gff_struct->{Fields}) { $gff_struct->{Fields}=[$gff_struct->{Fields}] }
		else { $gff_struct->{Fields}=[]}
		}
	 push @{$gff_struct->{Fields}},$gff_field;
	}
  }
  else {
	my $gff_field=reref_gff_field($ungff_struct->{Fields});
	$gff_struct->{Fields}=$gff_field;
  }
  return $gff_struct;
}
sub reref_gff_field {
  my $ungff_field=shift;
  return [] unless (ref $ungff_field) eq 'HASH';  # an empty struct
  my $gff_field=Bioware::GFF::Field->new();
  $gff_field->{FieldIndex}=$ungff_field->{FieldIndex};
  $gff_field->{Label}=$ungff_field->{Label};
  $gff_field->{Type}=$ungff_field->{Type};
  if ($ungff_field->{Type} == 12 ) {
    my $gff_locstring=Bioware::GFF::CExoLocString->new();
    $gff_locstring->{'StringRef'}=$ungff_field->{Value}{StringRef};
    my @gff_substrings;
    if (ref $ungff_field->{Value}{Substrings} eq 'ARRAY') {
      for my $substring (@{$ungff_field->{Value}{Substrings}}) {
	     my $gff_substring=reref_gff_substring($substring);
         push @gff_substrings, $gff_substring;
      }
	}
    $gff_locstring->{Substrings}=\@gff_substrings;
    $gff_field->{Value}=$gff_locstring;
  }
  elsif ($ungff_field->{Type} == 14) {
    $gff_field->{Value}=reref_gff_struct($ungff_field->{Value});
  }
  elsif ($ungff_field->{Type} == 15) {
    if (ref $ungff_field->{Value} eq 'ARRAY') {
        my @gff_structs;
        for my $ungff_struct (@{$ungff_field->{Value}}) { # has multiple structs
          my $gff_struct=reref_gff_struct($ungff_struct);
          push @gff_structs,$gff_struct;
        }
        $gff_field->{Value}=\@gff_structs;
    }
    elsif (ref $ungff_field->{Value} eq 'HASH') { # has a single struct
        my $gff_struct=reref_gff_struct($ungff_field->{Value});
        $gff_field->{Value}=$gff_struct;
    }
    else { # has no structs
        $gff_field->{Value}=[];
    }
  }
  else {
    $gff_field->{Value}=$ungff_field->{Value};
  }
  return $gff_field;
}
sub reref_gff_substring {
  my $substring=shift;
  my $gff_substring=Bioware::GFF::CExoLocSubString->new();
  $gff_substring->{StringID}=$substring->{StringID};
  $gff_substring->{Value}=$substring->{Value};
  return $gff_substring;
}
sub last_chance_to_save {
    #used to give the user one last chance to save their work before
    #closing the current GFF object
    if ($gff_dirty_flag) {
        my $d = $mw->DialogBox(-title => "Save Changes?",-buttons => ["Save", "Save As...","Close"]);
        $d->add('Label', -text=>"Do you wish to save changes to $current_gffname?")->pack;
        my $answer=$d->Show();
        if ($answer eq 'Save') {
            save_gff();
        }
        elsif ($answer eq 'Save As...') {
            save_as_gff();
        }
    }
}
sub expand_tree { #this will open all nodes
    my $treeitem=shift;
    if ($treeitem) {
	$tree->open($treeitem);
    }
    else {
	$treeitem='#';
	while ($treeitem=$tree->infoNext($treeitem)) {
	    $tree->open($treeitem);
	}
	$tree->open('#');
    }
}
sub fold_tree{ #this will close all nodes
    
    my $treeitem=shift;
    if ($treeitem) {
	$tree->close($treeitem);
    }
    else {
	$treeitem='#';
	while ($treeitem=$tree->infoNext($treeitem)) {
	    $tree->close($treeitem);
	}
    }
}
sub apply_changes {
    #user has clicked something, any changes made to the last treeitem should be recorded first
    if ($lasttreeitem) {
		my $lastdata=$tree->info('data',$lasttreeitem);
		if (ref $lastdata eq 'Bioware::GFF::Struct') {
		    $lastdata->{ID}=$txtID_var;
			$tree->entryconfigure($lasttreeitem,-text=>"[STRUCT ID: ".$lastdata->{ID}."]");
		}
		elsif (ref $lastdata eq 'Bioware::GFF::Field') {
		    $lastdata->{Label}=$txtLabel_var;
		    $lastdata->{Type}=$beType_varix;
		    if ($lastdata->{Type} < 12) {
				$lastdata->{Value}=$txtValue_var[0];
				$tree->delete('offsprings',$lasttreeitem);
		    }
		    elsif ($lastdata->{Type} == FIELD_CEXOLOCSTRING) {
				$lastdata->{Value}{StringRef}=$txtStringRef_var;
				#$tree->delete('offsprings',$lasttreeitem);
		    }
		    elsif ($lastdata->{Type} == FIELD_BINARY) {
				$tree->delete('offsprings',$lasttreeitem);
		    }
		    elsif ($lastdata->{Type} == FIELD_STRUCT) {
			if (ref $lastdata->{Value} ne 'Bioware::GFF::Struct') {
					$lastdata->{Value}=Bioware::GFF::Struct->new('ID'=>0);
					$tree->delete('offsprings',$lasttreeitem);
					$tree->add($lasttreeitem.'#S0',-text=>"[STRUCT ID: 0]",-data=>$lastdata->{Value});
			}
		    }
		    elsif ($lastdata->{Type} == FIELD_LIST) {
				$lastdata->{Value}=[] unless (ref $lastdata->{Value} eq 'ARRAY');
		    }
		    elsif ($lastdata->{Type} == FIELD_ORIENTATION) {
				$lastdata->{Value} = [	map	{unpack('f',pack('f',$_))} @txtValue_var];
		    }
		    elsif ($lastdata->{Type} == FIELD_POSITION) {
				$lastdata->{Value} = [map {unpack('f',pack('f',$_))} @txtValue_var[0..2]];
		    }
		    elsif ($lastdata->{Type} == FIELD_STRREF) {
				$lastdata->{Value}=$txtStringRef_var;
		    }

		    my $branchtext=$lastdata->{Label} . " [Type: " . $field_types[$lastdata->{Type}] . "] ";
			if ($lastdata->{Type} < 12) {
				$branchtext .= "Value: ".$lastdata->{Value}
			}
			if ($lastdata->{Type} == FIELD_CEXOLOCSTRING) {
				$branchtext .= "StringRef: ".$lastdata->{Value}{StringRef}
			}
			if ($lastdata->{Type} == FIELD_STRREF) {
				$branchtext .= "StringRef: ".$lastdata->{Value};
			}
			if ($lastdata->{Type} == FIELD_LIST) {
				my $structrefs=$lastdata->{Value};
				if (ref $structrefs eq 'ARRAY') {
				  $branchtext .= " Structs: " . scalar @$structrefs;
				}
				elsif (ref $structrefs eq 'Bioware::GFF::Struct') {
					$branchtext .= " Structs: 1";
				}
				else {
					$branchtext .= " Structs: 0";
				}
		    }

			$tree->entryconfigure($lasttreeitem,-text=>$branchtext);

		}
		elsif (ref $lastdata eq 'Bioware::GFF::CExoLocSubString') {
		    my $us=Unicode::String->new($txtLocString->get('1.0','end'));
			$lastdata->{Value}=$us->latin1;
			chomp $lastdata->{Value};
			my %rev_lang=reverse %language_types;
			$lastdata->{StringID}=$rev_lang{$beLanguage_var}*2+$beGender_varix;
			$lasttreeitem=~/T(\d+)$/;
			my $ss_treetext='[LocalString '.$1.'] ('.$beGender_var.' '.$beLanguage_var.') ';
			if (length ($lastdata->{Value} > 20)) {
				$ss_treetext .= substr($lastdata->{Value},0,20).'...';
			}
			else {
				$ss_treetext .= $lastdata->{Value}
			}
			$tree->entryconfigure($lasttreeitem,-text=>$ss_treetext);
		}
	}
}
sub delete_node {
	my $treeitem=shift;
	my $next_treeitem=$treeitem;
	{
	$next_treeitem=$tree->infoNext($next_treeitem);
	redo if $next_treeitem=~/^$treeitem#/; #we don't want children
	}

	$next_treeitem=$tree->infoPrev($treeitem) unless $next_treeitem;
	if ($treeitem=~/#S\d+$/) { delete_struct_node($treeitem) }
	elsif ($treeitem=~/#F\d+$/) { delete_field_node($treeitem)  }
	elsif ($treeitem=~/#T\d+$/) { delete_substring_node($treeitem) }

	return unless $tree->infoExists($next_treeitem);
	#now select next node
    $tree->selectionClear();
    $tree->selectionSet($next_treeitem);

    #now see the node
    $tree->see($next_treeitem);

    #spawn widgets ala left click
    tree_click($next_treeitem);


}
sub search_by_label {
	my $db=$mw->DialogBox(-title=>'Search By Label',-buttons=>['Search','Cancel'],-default_button=>'Search');
	$db->add('Label',-text=>"Enter Label to Search")->pack;
	my $search_term;
	my $ent=$db->add('LabEntry',-textvariable=>\$search_term,-background=>'white',-width=>16,-label=>'Label:',-labelPack=>[-side=>'left'])->pack;
	$ent->focus();
	my $response=$db->Show();
	return if $response eq 'Cancel';
	return unless $search_term;
        
        # warning line noise ahead:
        # just grabbing the label_lookup hash from the gff
        # and seaching it for our search term
        # and then grabbing the index and using it to lookup our treebranch
#        my @found_branches = map { /(.*?)\)/;  $gff->{tree_lookup}{$1}; }
#                             grep { /\).*$search_term/ }
#                             values %{$gff->{label_lookup}};
#                             
#        #this is a funky sort technique needed to search down the tree (rather than into the tree)
#        @found_branches=sort{
#            my @a_chunks = map { /(\d+)/; my $z=$1; $z;} (split /#/,$a);
#            my @b_chunks =map { /(\d+)/; my $z=$1; $z;} (split /#/,$b);
#            my $x;
#            while (($x==0) && ((scalar @a_chunks)||(scalar @b_chunks))) {
#                $x= (shift @a_chunks <=> shift @b_chunks);
#            }
#            return $x;
#        } @found_branches;        
#        while (my $found_branch = shift @found_branches) {
#            $tree->autosetmode;
#	    $tree->show('entry',$found_branch);                       #unfold this treeitem
#	    $tree->selectionClear();                              #clear previous selection
#	    $mw->idletasks;
#	    $tree->selectionSet($found_branch);                       #set new selection
#	    $mw->idletasks;
#	    $tree->see($found_branch);                                #make it visible
#	    tree_click($found_branch);                                #click it
#	    my $db=$mw->DialogBox(-title=>'Find Next?',-buttons=>['Next','Cancel'],-default_button=>'Next');
#	    $db->add('Label',-text=>"Find Next label of $search_term?")->pack;
#	    my $response=$db->Show();
#	    return if $response eq 'Cancel';
#        }
#	$msgbox->Call(0,"All Labels have been searched.","Search complete",0);
#	return;
	
    my $treeitem=($tree->infoSelection)[0];
	$treeitem='#S0' unless ($treeitem);
	$treeitem='#S0' if $treeitem eq '#';
	my $firsttreeitem;
	$treeitem=$tree->infoPrev($treeitem);
	{
		while ($treeitem=$tree->infoNext($treeitem)) {
			if ($treeitem eq $firsttreeitem) {
				$msgbox->Call(0,"All Labels have been searched.","Search complete",0);
				return;
			}
	        $firsttreeitem = $treeitem unless $firsttreeitem;
			next unless $treeitem =~/.*F\d+$/;                    #only search Fields
			my $dataitem=$tree->infoData($treeitem);              #get the Field
			next unless $dataitem->{Label} =~ /$search_term/;     #check the Label, loop if not what we want
			my $parent=$tree->infoParent($treeitem);
			while ($parent) {
				$tree->open($parent);
				$parent=$tree->infoParent($parent);
			}
			$tree->autosetmode;
			$tree->show('entry',$treeitem);                       #unfold this treeitem
			$tree->selectionClear();                              #clear previous selection
			$mw->idletasks;
			$tree->selectionSet($treeitem);                       #set new selection
			$mw->idletasks;
			$tree->see($treeitem);                                #make it visible
	 		tree_click($treeitem);                                #click it
			#find next?
		    my $db=$mw->DialogBox(-title=>'Find Next?',-buttons=>['Next','Cancel'],-default_button=>'Next');
		    $db->add('Label',-text=>"Find Next label of $search_term?")->pack;
			my $response=$db->Show();
			return if $response eq 'Cancel';
		}
		$treeitem='#';  #search back at the top
		redo;
	}
}
sub search_by_value {
	my $db=$mw->DialogBox(-title=>'Search By Value',-buttons=>['Search','Cancel'],-default_button=>'Search');
	$db->add('Label',-text=>"Enter Value to Search")->pack;
	my $search_term;
	my $ent=$db->add('LabEntry',-textvariable=>\$search_term,-background=>'white',-width=>16,-label=>'Value:',-labelPack=>[-side=>'left'])->pack;
	$ent->focus();
	my $response=$db->Show();
	return if $response eq 'Cancel';
	return if $search_term==undef;
    my $treeitem=($tree->infoSelection)[0];
	$treeitem='#S0' unless ($treeitem);
	$treeitem='#S0' if $treeitem eq '#';
	my $firsttreeitem;
	$treeitem=$tree->infoPrev($treeitem);
	{
		while ($treeitem=$tree->infoNext($treeitem)) {
			if ($treeitem eq $firsttreeitem) {
				$msgbox->Call(0,"All Values have been searched.","Search complete",0);
				return;
			}
	        $firsttreeitem = $treeitem unless $firsttreeitem;
			next unless $treeitem =~/.*F\d+$/;                    #only search Fields
			my $dataitem=$tree->infoData($treeitem);              #get the Field
			next if $dataitem->{Type}==FIELD_CEXOLOCSTRING;
			next if $dataitem->{Type}==FIELD_BINARY;
			next if $dataitem->{Type}==FIELD_STRUCT;
			next if $dataitem->{Type}==FIELD_LIST;
			if ($dataitem->{Type} < FIELD_CEXOSTRING) {
				next unless $dataitem->{Value} == $search_term; #check simple value, loop if not what we want
			}
			elsif (($dataitem->{Type} == FIELD_CEXOSTRING) ||
			       ($dataitem->{Type} == FIELD_RESREF)) {
				next unless $dataitem->{Value} =~ /$search_term/; #check text value
			}
			elsif (($dataitem->{Type}== FIELD_ORIENTATION) ||
			       ($dataitem->{Type}== FIELD_POSITION)) {
			    my @v_arr=@{$dataitem->{Value}};
				next unless grep{$_==$search_term} @v_arr;  #check each value in array of values
			}
			my $parent=$tree->infoParent($treeitem);
			while ($parent) {
				$tree->open($parent);
				$parent=$tree->infoParent($parent);
			}
			$tree->autosetmode;

			$tree->show('entry',$treeitem);                       #unfold this treeitem
			$tree->selectionClear();                              #clear previous selection
			$mw->idletasks;
			$tree->selectionSet($treeitem);                       #set new selection
			$mw->idletasks;
			$tree->see($treeitem);                                #make it visible
	 		tree_click($treeitem);                                #click it
			#find next?
		    my $db=$mw->DialogBox(-title=>'Find Next?',-buttons=>['Next','Cancel'],-default_button=>'Next');
		    $db->add('Label',-text=>"Find Next value of $search_term?")->pack;
			my $response=$db->Show();
			return if $response eq 'Cancel';
		}
		$treeitem='#';  #search back at the top
		redo;
	}

}
sub search_by_stringref {
	my $db=$mw->DialogBox(-title=>'Search By StringRef',-buttons=>['Search','Cancel'],-default_button=>'Search');
	$db->add('Label',-text=>"Enter StringRef to Search")->pack;
	my $search_term;
	my $ent=$db->add('LabEntry',-textvariable=>\$search_term,-background=>'white',-width=>8,-label=>'StringRef:',-labelPack=>[-side=>'left'])->pack;
	$ent->focus();
	my $response=$db->Show();
	return if $response eq 'Cancel';
	return if $search_term==undef;
    my $treeitem=($tree->infoSelection)[0];
	$treeitem='#S0' unless ($treeitem);
	$treeitem='#S0' if $treeitem eq '#';
	my $firsttreeitem;
	$treeitem=$tree->infoPrev($treeitem);
	{
		while ($treeitem=$tree->infoNext($treeitem)) {
			if ($treeitem eq $firsttreeitem) {
				$msgbox->Call(0,"All Values have been searched.","Search complete",0);
				return;
			}
	        $firsttreeitem = $treeitem unless $firsttreeitem;
			next unless $treeitem =~/.*F\d+$/;                    #only search Fields
			my $dataitem=$tree->infoData($treeitem);              #get the Field
			
                        next unless (($dataitem->{Type}==FIELD_CEXOLOCSTRING) || ($dataitem->{Type}==FIELD_STRREF));
                        if ($dataitem->{Type}==FIELD_CEXOLOCSTRING) {
                            next unless $dataitem->{Value}{StringRef} == $search_term;
                        }
                        elsif ($dataitem->{Type}==FIELD_STRREF) {
                            next unless $dataitem->{Value} == $search_term;
                        }
			my $parent=$tree->infoParent($treeitem);
			while ($parent) {
				$tree->open($parent);
				$parent=$tree->infoParent($parent);
			}
			$tree->autosetmode;

			$tree->show('entry',$treeitem);                       #unfold this treeitem
			$tree->selectionClear();                              #clear previous selection
			$mw->idletasks;
			$tree->selectionSet($treeitem);                       #set new selection
			$mw->idletasks;
			$tree->see($treeitem);                                #make it visible
	 		tree_click($treeitem);                                #click it
			#find next?
		    my $db=$mw->DialogBox(-title=>'Find Next?',-buttons=>['Next','Cancel'],-default_button=>'Next');
		    $db->add('Label',-text=>"Find Next StringRef of $search_term?")->pack;
			my $response=$db->Show();
			return if $response eq 'Cancel';
		}
		$treeitem='#';  #search back at the top
		redo;
	}
}
sub go_to {
    my $db=$mw->DialogBox(-title=>'Go to',-buttons=>['Go To','Cancel'],-default_button=>'Go To');
    $db->add('Label',-text=>"Go to Struct# (S0..S$max_structindex)\nor Field#(F0..F$max_fieldindex)")->pack;
    my $goto_term;
    my $ent=$db->add('LabEntry',-textvariable=>\$goto_term,-background=>'white',-width=>16,-label=>'Text:',-labelPack=>[-side=>'left'])->pack;
    $ent->focus();
    my $response=$db->Show();
    return if $response eq 'Cancel';
    $goto_term=uc($goto_term);
    return unless $goto_term =~ /([SF])(\d+)/;
    my $s_or_f=$1;
    my $index=$2;
    if ($s_or_f eq 'S' ) {
        return if $index>$max_structindex;
    }
    else {
        return if $index>$max_fieldindex;
    }
    return unless exists $gff->{tree_lookup}{$goto_term};
    my $treeitem = $gff->{tree_lookup}{$goto_term};
    $tree->autosetmode;
    $tree->show('entry',$treeitem);                       #unfold this treeitem
    $tree->selectionClear();                              #clear previous selection
    $mw->idletasks;
    $tree->selectionSet($treeitem);                       #set new selection
    $mw->idletasks;
    $tree->see($treeitem);                                #make it visible
    tree_click($treeitem);                                #click it
}
sub search_by_text {
	my $db=$mw->DialogBox(-title=>'Search By Text',-buttons=>['Search','Cancel'],-default_button=>'Search');
	$db->add('Label',-text=>"Enter Text to Search")->pack;
	my $search_term;
	my $ent=$db->add('LabEntry',-textvariable=>\$search_term,-background=>'white',-width=>16,-label=>'Text:',-labelPack=>[-side=>'left'])->pack;
	$ent->focus();
	my $response=$db->Show();
	return if $response eq 'Cancel';
	return unless $search_term;
    my $treeitem=($tree->infoSelection)[0];
	$treeitem='#S0' unless ($treeitem);
	$treeitem='#S0' if $treeitem eq '#';
	my $firsttreeitem;
	$treeitem=$tree->infoPrev($treeitem);
	{
		while ($treeitem=$tree->infoNext($treeitem)) {
			if ($treeitem eq $firsttreeitem) {
                            if ($tlkfile_in_use==0) {
                                $msgbox->Call(0,"All Values have been searched.\nNote: TLK file is not currently used\nPress Ctrl+T to use TLK file.","Search complete",0);
                                
                            }
                            else {
				$msgbox->Call(0,"All Values have been searched.","Search complete",0);
                            }
			    return;
			}
	        $firsttreeitem = $treeitem unless $firsttreeitem;
			next if $treeitem=~/.*S\d+$/;
			if ($treeitem=~/.*T\d+$/)  {                          #Cexolocsubstring search
				my $dataitem=$tree->infoData($treeitem);              #get the substring
				next unless $dataitem->{Value} =~ /$search_term/;
			}
		    else {
			my $dataitem=$tree->infoData($treeitem);              #get the Field
		        next unless (($dataitem->{Type} == FIELD_CEXOSTRING) ||
			       ($dataitem->{Type} == FIELD_RESREF)  ||
                               ($dataitem->{Type} == FIELD_STRREF)  ||
                               ($dataitem->{Type} == FIELD_CEXOLOCSTRING));
				if ($dataitem->{Type} == FIELD_STRREF) {
                                    next unless get_tlk_string($dataitem->{Value}) =~ /$search_term/;
                                }
                                elsif ($dataitem->{Type}== FIELD_CEXOLOCSTRING) {
                                    next unless get_tlk_string($dataitem->{Value}{StringRef}) =~ /$search_term/;
                                }
                                else {
                                    next unless $dataitem->{Value} =~ /$search_term/; #check text value
                                }
			}
			my $parent=$tree->infoParent($treeitem);
			while ($parent) {
				$tree->open($parent);
				$parent=$tree->infoParent($parent);
			}
			$tree->autosetmode;

			$tree->show('entry',$treeitem);                       #unfold this treeitem
			$tree->selectionClear();                              #clear previous selection
			$mw->idletasks;
			$tree->selectionSet($treeitem);                       #set new selection
			$mw->idletasks;
			$tree->see($treeitem);                                #make it visible
	 		tree_click($treeitem);                                #click it
			#find next?
		    my $db=$mw->DialogBox(-title=>'Find Next?',-buttons=>['Next','Cancel'],-default_button=>'Next');
		    $db->add('Label',-text=>"Find Next value of $search_term?")->pack;
			my $response=$db->Show();
			return if $response eq 'Cancel';
		}
		$treeitem='#';  #search back at the top
		redo;
	}

}
sub read_mru_list {
    $mrulist=List::MRU->new(max=>$mrulist_max);
    for (my $i=$mrulist_max; $i; $i--) {
      if (my $temp=iniReadString('MRU',$i,'','kgff.ini')) {
          $mrulist->add($temp) if (-e $temp);  #the "if (-e $temp)"  was added v1.2.0
      }
    }
	regenerate_menubar();
}
sub regenerate_menubar {
    return unless $menubar;
	# store current states of menu items for each menubar
	my %menubar_states;
	for (my $i=0; $i<= $menubar->index('end'); $i++) {
	  my $this_cascade=$menubar->entrycget($i,-menu);
	  my $this_cascade_label=$menubar->entrycget($i,-label);
	  for (my $ii=0; $ii<=$this_cascade->index('end'); $ii++) {
	    unless ($this_cascade->type($ii) eq 'separator') {
	      my $this_cascade_entry_label=$this_cascade->entrycget($ii,-label);
		  my $this_cascade_entry_state=$this_cascade->entrycget($ii,-state);
		  $menubar_states{$this_cascade_label}{$this_cascade_entry_label}=$this_cascade_entry_state;
		}
	  }
	}
	# then recreate the menubar
	$menubar->delete('File');
	$menubar->delete('Edit');
	$menubar->delete('View');
	$menubar->delete('Tools');
	$menubar->cascade(-label=>'~File',-menuitems=>mnuFile,-tearoff=>0);
	$menubar->cascade(-label=>'~Edit',-menuitems=>mnuEdit,-tearoff=>0);
	$menubar->cascade(-label=>'~Tools',-menuitems=>mnuTools,-tearoff=>0);
	$menubar->cascade(-label=>'~View',-menuitems=>mnuView,-tearoff=>0);
	# and restore old states
	for my $cascade_label (keys %menubar_states) {
	  my $cascade=$menubar->entrycget($cascade_label,-menu);
	  for my $item_label (keys %{$menubar_states{$cascade_label}}) {
	    unless ($item_label=~/^~?\d+\s/) {
	      $cascade->entryconfigure($item_label,-state=>$menubar_states{$cascade_label}{$item_label});
		}
	  }
	}
}
sub write_mru_list {
    for (my $i=1; $i<=$mrulist->count; $i++) {
       iniWriteString('MRU',$i,($mrulist->list)[$i-1],'kgff.ini');
    }
}

sub update_mru {
    my $latest=shift;
    $mrulist->add($latest);
	regenerate_menubar();
    write_mru_list();
}
sub set_mru_length {
  my $ent_var=$mrulist_max;
  my $db=$mw->DialogBox(-title=>'Set MRU List Length',-buttons=>['Ok']);
  $db->add('Label',-text=>'Maximum number of most recently used files to show')->pack;
  my $ent=$db->add('Entry',-textvariable=>\$ent_var,-background=>'white',-width=>4)->pack;
  $ent->focus();
  $db->Show();
  if ( ($ent_var +=0) > 0) {
    $mrulist_max=$ent_var;
    iniWriteString('Settings','MRUMax',$mrulist_max,'kgff.ini');
    read_mru_list();
  }
}
sub get_window_state {
  my $hwnd=hex($mw->frame);
  my $ret=$IsZoomed->Call($hwnd);
  iniWriteString('Main','Maximized',$ret,'kgff.ini');
  $ret=$IsIconic->Call($hwnd);
  iniWriteString('Main','Minimized',$ret,'kgff.ini');
}
sub set_window_state {
  my $win_version=$GetVersion->Call();
  $win_version = $win_version & hex("FFFF");
  $win_version = ($win_version & hex("FF")) + (($win_version>>8)/100);
  return if ($win_version>5.0);

  my $min=iniReadString('Main','Minimized','','kgff.ini');
  my $max=iniReadString('Main','Maximized','','kgff.ini');
  my $hwnd=hex($mw->frame);

  $ShowWindow->Call($hwnd,10); #Default
  $UpdateWindow->Call($hwnd);

  if ($min) {
	$ShowWindow->Call($hwnd,2);
  }
  elsif ($max) {
	$ShowWindow->Call($hwnd,3);
  }
}
sub change_to_v32_lang {
    %language_types=%lang_v32_types;
    $beLanguage->Subwidget('slistbox')->delete(0,'end');#configure('-choices')=[];
    $beLanguage->Subwidget('slistbox')->insert(0,map {$language_types{$_}} sort {$a<=>$b} keys %language_types);
    iniWriteString('Main','Language ID Definition',$v32_lang,'kgff.ini');
    my $treeitem=($tree->infoSelection)[0];
    refresh_tree();
}
sub change_to_v33_lang {
    %language_types=%lang_v33_types;
    $beLanguage->Subwidget('slistbox')->delete(0,'end');#configure('-choices')=[];
    $beLanguage->Subwidget('slistbox')->insert(0,map {$language_types{$_}} sort {$a<=>$b} keys %language_types);
    iniWriteString('Main','Language ID Definition',$v32_lang,'kgff.ini');
    my $treeitem=($tree->infoSelection)[0];
    refresh_tree();

}